// ============================================================================
// filesBD.c
// ----------------------------------------------------------------------------
// (c) 2006-2023 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// loader for BD game files
// ----------------------------------------------------------------------------
// credits to Arno Weber - who help a lot
// ============================================================================

#include <ctype.h>
#include <sys/stat.h>
#include <math.h>

#include "libgame/libgame.h"
#include "libgame/misc.h"

#include "files.h"
#include "filesBD.h"
#include "filesBDCFF.h"
#include "filesGAMEBD.h"
#include "init.h"
#include "tools.h"
#include "tape.h"
#include "main.h"
#include "libgame/rndapi.h"		/*#HAG#VC#*/


/*=================================================================*/
/* data structure of BD level files */
/*--------------------------------------------------*/
#define BD_FILE_MAXLEN  1024 

#define BD_CAVE_WIDTH            40
#define BD_CAVE_HEIGHT           22
#define BD_INTERMISSION_WIDTH    20
#define BD_INTERMISSION_HEIGHT   12



/*--------------------------------------------------*/
static int chint (char ch) {
  int ia;
  ia = ch;
  if (ia < 0) {
    ia += 256;
  }
  return (ia);
}


/*--------------------------------------------------*/
static int bcd1int (char ch) {
  int ia;
  ia = ch;
  if (ia < 0) {
    ia += 256;
  }
  ia = ia % 16; 
  if (ia > 9) { fprintf (stderr, "bcd1int ia > 9 \n"); }

  return (ia);
}


/*--------------------------------------------------*/
static int bcd2int (char ch[2]) {
  int ia,ib;
  ia = ch[0];
  if (ia < 0) {
    ia += 256;
  }
  ia = ia % 16; 
  if (ia > 9) { fprintf (stderr, "bcd2int ia > 9 \n"); }

  ib = ch[1];
  if (ib < 0) {
    ib += 256;
  }
  ib = ib % 16;
  if (ib > 9) { fprintf (stderr, "bcd2int ib > 9 \n"); }

  return (ia*10 + ib);
}


/*--------------------------------------------------*/
static int bcd3int (char ch[3]) {
  int ia,ib,ic;

  ia = ch[0];
  if (ia < 0) {
    ia += 256;
  }
  ia = ia % 16;
  if (ia > 9) { fprintf (stderr, "bcd3int ia > 9 \n"); }

  ib = ch[1];
  if (ib < 0) {
    ib += 256;
  }
  ib = ib % 16;
  if (ib > 9) { fprintf (stderr, "bcd3int ib > 9 \n"); }

  ic = ch[2];
  if (ic < 0) {
    ic += 256;
  }
  ic = ic % 16;
  if (ic > 9) { fprintf (stderr, "bcd3int ic > 9 \n"); }

  return (ia*100 + ib*10 + ic);
}


/*--------------------------------------------------*/
static int color_to_int (char ch, int mask)
{
  int ia;

  ia = ch;
  if (ia < 0) {
    ia += 256;
  }

  ia = ia & mask ;   // mask
  ia = ia & 0x0f ;   // common mask

  return (ia);
}


/*======================*/
/*=== all for BD64 ===> */
/*======================*/

typedef struct bd_bd64_data_s {
  char      fileid[2];
  char      ldata[440];
  char      cavedelay;
  char      unused01;
  char      cavetime;
  char      unused02;
  char      diamneeded;
  char      unused03;
  char      diamvalue;
  char      unused04;
  char      diambonus;
  char      unused05;
  char      slimperm;
  char      unused06;
  char      amoebatime;
  char      unused07;
  char      magwalltime;
  char      unused08;
  char      unknown1[18];
  char      flags;
  char      color1;
  char      unused09;
  char      color2;
  char      unused10;
  char      color3;
  char      unused11;
  char      color4;
  char      unused12;
  char      color5;
  char      unknown2[12];
} bd_bd64_data_t;

static char bd_bd64_fileid_head[2]         = { 0x00, 0xBE };
static char bd_bd64_flagid_intermission[1] = { 0x3C };


/*=======================*/
/*=== all for ATARI ===> */
/*=======================*/

typedef struct bd_atari_data_s {
  char      fileid[4];
  char      cavename[12];
  char      ldata[440];
  char      unknown01[16];
  char      cavedelay[2];
  char      cavetime[3];
  char      diamneeded[2];
  char      diamvalue[2];
  char      diambonus[2];
  char      slimperm;
  char      amoebatime[3];
  char      magwalltime[3];
  char      flags;
  char      color1;
  char      unused01;
  char      color2;
  char      unused02;
  char      color3;
  char      unused03;
  char      color4;
  char      unused04;
  char      color5;
  char      unknown03[4];
} bd_atari_data_t;

static char bd_atari_fileid_head[4] = { 0xAB, 0xAB, 0xAB, 0xAB };
static char bd_atari_flagid_intermission[1] = { 0x3C };


/*======================*/
/*=== all for WLC ===> */
/*======================*/

typedef struct bd_wlc_data_s {
  char      fileid[5];
  char      ldata[880];
  char      filler[0x10-5];     // parastart after filler // 0x00 - 0x1F
  char      caveno;             // 00 
  char      magwalltime;        // 01 // amoeba slow growth for: x seconds
  char      diamvalue;          // 02
  char      diambonus;          // 03
  char      randseed[5];        // 04,05,06,07,08
  char      diamneeded[5];      // 09,10,11,12,13
  char      cavetime[5];        // 14,15,16,17,18
  char      colorbg1;           // 19 // background color 1
  char      colorbg2;           // 20 // background color 2
  char      colorfg;            // 21 // foreground color
  char      unused[2];          // 22,23
  char      randobjects[4];     // 24,25,26,27
  char      randvalues[4];      // 28,29,30,31
} bd_wlc_data_t;

static char bd_wlc_fileid_head[5]         = { 0x00, 0xC1, 'W', 'L', 'C' };
// intermission is cave Q,R,S,T


/*======================*/
/*=== all for WL2 ===> */
/*======================*/

typedef struct bd_wl2_data_s {
  char      fileid[5];
  char      ldata[880];
  char      filler[0x10-5];     // parastart after filler // raw: 0x00 - 0x19 -- with colors: 0x00 - 0x1F // command start at 0x20
  char      magwalltime;        // 00 // parastart // amoeba slow growth for: x seconds
  char      diamvalue;          // 01
  char      diambonus;          // 02
  char      cavetime[5];        // 03,04,05,06,07
  char      diamneeded[5];      // 08,09,10,11,12
  char      randseed[5];        // 13,14,15,16,17
  char      randvalues[4];      // 18,19,20,21
  char      randobjects[4];     // 22,23,24,25
  char      colorbg1;           // 26
  char      colorbg2;           // 27
  char      colorfg;            // 28
  char      unknown01;          // 29
  char      unknown02;          // 30
  char      filler01;           // 31
} bd_wl2_data_t;

static char bd_wl2_fileid_head[5]         = { 0x00, 0xC1, 'W', 'L', '2' };
// intermission is cave Q,R,S,T


/*======================*/
/*=== all for OTO ===> */
/*======================*/

typedef struct bd_oto_data_s {
  char      fileid[5];
  char      ldata[880];
  char      para[62];           // parameter
} bd_oto_data_t;

static char bd_oto_fileid_head[5]         = { 0x00, 0xC0, 'O', 'T', 'O' };
// intermission is cave Q,R,S,T


/*=================================================================*/

/*====================================================================*/

static int bd_atari_map(int ch) {
  int el;

  switch (ch) {
    case  0 : el = EL_BD_ROCK;                   /* boulder */
              break;
    case  1 : el = EL_BD_DIAMOND;                /* diamond */
              break;
    case  2 : el = EL_BD_MAGIC_WALL;             /* magic wall */
              break;
    case  3 : el = EL_BD_WALL;                   /* brick wall */
              break;
    case  4 : el = EL_STEELWALL;                 /* steel wall */
              break;
    case  5 : el = EL_BD_EXPANDABLE_WALL;        /* expand wall */
              break;
    case  6 : el = EL_PLAYER_1;                  /* rockford */
              break;
    case  7 : el = EL_SAND;                      /* dirt */
              break;
    case  8 : el = EL_BD_FIREFLY;                /* firefly */
              break;
    case  9 : el = EL_BD_BUTTERFLY;              /* butterfly */
              break;
    case 10 : el = EL_BD_AMOEBA;                 /* amoeba */
              break;
    case 11 : el = EL_AMOEBA_WET;                /* slime */
              break;
    case 12 : el = EL_EM_STEEL_EXIT_CLOSED;      /* hidden out */
              break;
    case 13 : el = EL_EM_EXIT_CLOSED;            /* visible out */
              break;
    case 14 : el = EL_PLAYER_1;                  /* in */
              break;
    case 15 : el = EL_EMPTY;                     /* space */
              break;
    default : el = EL_STEEL_CHAR_QUESTION;   
              break;
  }
  return el;
}

static int bd_bd64_map(int ch) {
  int el;
  el = bd_atari_map(ch);  // same as atari
  return el;
}

/*----------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------------*/
static int bd_element_mapNibble(int ch, int x, int y, int lev, int elDefault) {
    int el;
    if (elDefault < 0) {
        elDefault = EL_STEEL_CHAR_QUESTION;
    }
    //----
    // normal elements 
    // nibble code convert to this                  // nibble
    /*----
    case  0x10 : el = EL_BD_ROCK;                   // 0x00 ( 0)  = boulder
                 break;
    case  0x14 : el = EL_BD_DIAMOND;                // 0x01 ( 1)  = diamond
                 break;
    case  0x03 : el = EL_BD_MAGIC_WALL;             // 0x02 ( 2) = magic wall
                 break;
    case  0x02 : el = EL_BD_WALL;                   // 0x03 ( 3) = brick wall
                 break;
    case  0x07 : el = EL_STEELWALL;                 // 0x04 ( 4) = steel wall
                 break;
    case  0x2A : el = EL_BD_EXPANDABLE_WALL;        // 0x05 ( 5) = expand wall
                 break;
    case  0x3C : el = EL_PLAYER_1;                  // 0x06 ( 6) = rockford (fake like wall)
                 break;
    case  0x01 : el = EL_SAND;                      // 0x07 ( 7) = dirt
                 break;
    case  0x08 : el = EL_BD_FIREFLY;                // 0x08 ( 8) = firefly
                 break;
    case  0x30 : el = EL_BD_BUTTERFLY;              // 0x09 ( 9) = butterfly
                 break;
    case  0x3A : el = EL_BD_AMOEBA;                 // 0x0A (10) = amoeba
                 break;
    case  0x3E : el = EL_AMOEBA_WET;                // 0x0B (11) = slime
                 break;
    case  0x3D : el = EL_EM_STEEL_EXIT_CLOSED;      // 0x0C (12) = hidden out
                 break;
    case  0x04 : el = EL_EM_EXIT_CLOSED;            // 0x0D (13) = visible out
                 break;
    case  0x25 : el = EL_PLAYER_1;                  // 0x0E (14) = in
                 break;
    case  0x00 : el = EL_EMPTY;                     // 0x0F (15) = space
                 break;
    ----*/
    switch(ch){
        case 0x10    : el = EL_BD_ROCK;                 break;      // retval = 'r';  break;  // 00 // ok // EL_BD_ROCK               // retval = 'o';  break;  // rock
        case 0x14    : el = EL_BD_DIAMOND;              break;      // retval = 'd';  break;  // 01 // ok // EL_BD_DIAMOND            // retval = '*';  break;  // diamond
        case 0x03    : el = EL_BD_MAGIC_WALL;           break;      // retval = 'm';  break;  // 02 // ok // EL_BD_MAGIC_WALL         // retval = 'm';  break;  // magicWall
        case 0x02    : el = EL_BD_WALL;                 break;      // retval = 'w';  break;  // 03 // ok // EL_BD_WALL               // retval = 'W';  break;  // wall
        case 0x07    : el = EL_STEELWALL;               break;      // retval = 'W';  break;  // 04 // ok // EL_STEELWALL             // retval = '#';  break;  // steelWall
        case 0x2A    : el = EL_BD_EXPANDABLE_WALL;      break;      // retval = 'E';  break;  // 05 //+ok+// EL_BD_EXPANDABLE_WALL    // retval = 'e';  break;  // expandWall
        case 0x3C    : el = EL_PLAYER_1;                break;      // retval = 'R';  break;  // 06 //+ok+// EL_PLAYER_1              // retval = 'r';  break;  // rockford (fake like wall)
        case 0x01    : el = EL_SAND;                    break;      // retval = '.';  break;  // 07 // ok // EL_SAND                  // retval = ':';  break;  // dirt
        case 0x08    : el = EL_BD_FIREFLY_1;            break;      // retval = 'q';  break;  // 08 // ok // EL_BD_FIREFLY_1          // retval = 'f';  break;  // firefly   (links)
        case 0x30    : el = EL_BD_BUTTERFLY_1;          break;      // retval = 'B';  break;  // 09 // ok // EL_BD_BUTTERFLY_1        // retval = 'b';  break;  // butterfly
        case 0x3A    : el = EL_BD_AMOEBA;               break;      // retval = 'a';  break;  // 0A // ok // EL_BD_AMOEBA             // retval = 'a';  break;  // amoeba
        case 0x3E    : el = EL_AMOEBA_WET;              break;      // retval = 'S';  break;  // 0B // ok // EL_AMOEBA_WET            // retval = 's';  break;  // slime
        case 0x3D    : el = EL_EM_STEEL_EXIT_CLOSED;    break;      // retval = 'H';  break;  // 0C // ok // EL_EM_STEEL_EXIT_CLOSED  // retval = 'H';  break;  // hiddenOut
        case 0x04    : el = EL_EM_EXIT_CLOSED;          break;      // retval = 'P';  break;  // 0D // ok // EL_EM_EXIT_CLOSED        // retval = 'V';  break;  // visibleOut
        case 0x25    : el = EL_PLAYER_1;                break;      // retval = 'X';  break;  // 0E // ok // EL_PLAYER_1              // retval = 'P';  break;  // player
        case 0x00    : el = EL_EMPTY;                   break;      // retval = ' ';  break;  // 0F // ok // EL_EMPTY                 // retval = ' ';  break;  // space (empty)
        default      : el = elDefault;                  break;      // retval = '?';  break;
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static int bd_element_mapFull(int ch, int x, int y, int lev, int elDefault) {
    int el;
    if (elDefault < 0) {
        elDefault = EL_STEEL_CHAR_QUESTION;
    }
    switch (ch) {
    //-------------------------------------------------------------------------------------
    // object codes here:
    // ->  https://www.elmerproductions.com/sp/peterb/rawCaveData.html#Object%20codes
    // char retval= '?';
    //-------------------------------------------------------------------------------------
    // elements
    case  0x00 : el = EL_EMPTY;                     // $00 Space
                 break;
    case  0x01 : el = EL_SAND;                      // $01 Dirt
                 break;
    case  0x02 : el = EL_BD_WALL;                   // $02 Brick wall
                 break;
    case  0x03 : el = EL_BD_MAGIC_WALL;             // $03 Magic wall
                 break;
    case  0x04 : el = EL_EM_EXIT_CLOSED;            // $04 Pre outbox (ie before it starts flashing)
                 break;
    case  0x05 : el = EL_EM_EXIT_OPEN;              // $05 Flashing outbox
                 break;
    case  0x06 : el = EL_STEEL_CHAR_0;              // $06 *    //-- not defined: "0"
                 break;
    case  0x07 : el = EL_STEELWALL;                 // $07 Steel wall
                 break;
    case  0x08 : el = EL_BD_FIREFLY_1;              // $08 Firefly position 1
                 break;
    case  0x09 : el = EL_BD_FIREFLY_2;              // $09 Firefly position 2
                 break;
    case  0x0A : el = EL_BD_FIREFLY_3;              // $0A Firefly position 3
                 break;
    case  0x0B : el = EL_BD_FIREFLY_4;              // $0B Firefly position 4
                 break;
    case  0x0C : el = EL_BD_FIREFLY_1;              // $0C Firefly position 1, scanned this frame
                 break;
    case  0x0D : el = EL_BD_FIREFLY_2;              // $0D Firefly position 2, scanned this frame
                 break;
    case  0x0E : el = EL_BD_FIREFLY_3;              // $0E Firefly position 3, scanned this frame
                 break;
    case  0x0F : el = EL_BD_FIREFLY_4;              // $0F Firefly position 4, scanned this frame
                 break;
    case  0x10 : el = EL_BD_ROCK;                   // $10 Boulder, stationary
                 break;
    case  0x11 : el = EL_BD_ROCK;                   // $11 Boulder, stationary, scanned this frame
                 break;
    case  0x12 : el = EL_BD_ROCK;                   // $12 Boulder, falling
                 break;
    case  0x13 : el = EL_BD_ROCK;                   // $13 Boulder, falling, scanned this frame
                 break;
    case  0x14 : el = EL_BD_DIAMOND;                // $14 Diamond, stationary
                 break;
    case  0x15 : el = EL_BD_DIAMOND;                // $15 Diamond, stationary, scanned this frame
                 break;
    case  0x16 : el = EL_BD_DIAMOND;                // $16 Diamond, falling
                 break;
    case  0x17 : el = EL_BD_DIAMOND;                // $17 Diamond, falling, scanned this frame
                 break;
    case  0x18 : el = EL_STEEL_CHAR_1;              // $18    //-- not defined: "1"
                 break;
    case  0x19 : el = EL_STEEL_CHAR_2;              // $19    //-- not defined: "2"
                 break;
    case  0x1A : el = EL_STEEL_CHAR_3;              // $1A    //-- not defined: "3"
                 break;
    case  0x1B : el = EL_EXPLOSION;                 // $1B Explode to Space stage 0
                 break;
    case  0x1C : el = EL_EXPLOSION;                 // $1C Explode to Space stage 1
                 break;
    case  0x1D : el = EL_EXPLOSION;                 // $1D Explode to Space stage 2
                 break;
    case  0x1E : el = EL_EXPLOSION;                 // $1E Explode to Space stage 3
                 break;
    case  0x1F : el = EL_EXPLOSION;                 // $1F Explode to Space stage 4
                 break;
    case  0x20 : el = EL_EXPLOSION;                 // $20 Explode to Diamond stage 0
                 break;
    case  0x21 : el = EL_EXPLOSION;                 // $21 Explode to Diamond stage 1
                 break;
    case  0x22 : el = EL_EXPLOSION;                 // $22 Explode to Diamond stage 2
                 break;
    case  0x23 : el = EL_EXPLOSION;                 // $23 Explode to Diamond stage 3
                 break;
    case  0x24 : el = EL_EXPLOSION;                 // $24 Explode to Diamond stage 4
                 break;
    case  0x25 : el = EL_PLAYER_1;                  // $25 Pre Rockford stage 1 (ie when inbox morphs into Rockford)
                 break;
    case  0x26 : el = EL_PLAYER_1;                  // $26 Pre Rockford stage 2
                 break;
    case  0x27 : el = EL_PLAYER_1;                  // $27 Pre Rockford stage 3
                 break;
    case  0x28 : el = EL_PLAYER_1;                  // $28 Pre Rockford stage 4
                 break;
    case  0x29 : el = EL_STEEL_CHAR_9;              // $29    //-- not defined: "9"
                 break;
    case  0x2A : el = EL_STEEL_CHAR_A;              // $2A    //-- not defined: "A"     // see replacement!! // expandWall
                 break;
    case  0x2B : el = EL_STEEL_CHAR_B;              // $2B    //-- not defined: "B"
                 break;
    case  0x2C : el = EL_STEEL_CHAR_C;              // $2C    //-- not defined: "C"
                 break;
    case  0x2D : el = EL_STEEL_CHAR_D;              // $2D    //-- not defined: "D"
                 break;
    case  0x2E : el = EL_STEEL_CHAR_E;              // $2E    //-- not defined: "E"
                 break;
    case  0x2F : el = EL_STEEL_CHAR_F;              // $2F    //-- not defined: "F"
                 break;
    case  0x30 : el = EL_BD_BUTTERFLY_1;            // $30 Butterfly position 1
                 break;
    case  0x31 : el = EL_BD_BUTTERFLY_2;            // $31 Butterfly position 2
                 break;
    case  0x32 : el = EL_BD_BUTTERFLY_3;            // $32 Butterfly position 3
                 break;
    case  0x33 : el = EL_BD_BUTTERFLY_4;            // $33 Butterfly position 4
                 break;
    case  0x34 : el = EL_BD_BUTTERFLY_1;            // $34 Butterfly position 1, scanned this frame
                 break;
    case  0x35 : el = EL_BD_BUTTERFLY_2;            // $35 Butterfly position 2, scanned this frame
                 break;
    case  0x36 : el = EL_BD_BUTTERFLY_3;            // $36 Butterfly position 3, scanned this frame
                 break;
    case  0x37 : el = EL_BD_BUTTERFLY_4;            // $37 Butterfly position 4, scanned this frame
                 break;
    case  0x38 : el = EL_PLAYER_1;                  // $38 Rockford?
                 break;
    case  0x39 : el = EL_PLAYER_1;                  // $39 Rockford?, scanned this frame
                 break;
    case  0x3A : el = EL_BD_AMOEBA;                 // $3A Amoeba
                 break;
    case  0x3B : el = EL_BD_AMOEBA;                 // $3B Amoeba, scanned this frame
                 break;
    case  0x3C : el = EL_STEEL_CHAR_G;              // $3C    //-- not defined: "G"     // see replacement!! // rockfordFake
                 break;
    case  0x3D : el = EL_STEEL_CHAR_H;              // $3D    //-- not defined: "H"     // see replacement!! // hiddenOut
                 break;
    case  0x3E : el = EL_STEEL_CHAR_I;              // $3E *  //-- not defined: "I"     // see replacement!! // slime
                 break;
    case  0x3F : el = EL_STEEL_CHAR_J;              // $3F *  //-- not defined: "J"
                 break;
    default    : el = elDefault;
                 break;
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static int bd_element_mapReplacement(int ch, int x, int y, int lev, int elDefault) {
    int el;
    if (elDefault < 0) {
        elDefault = EL_STEEL_CHAR_QUESTION;
    }
    // determine from caves // replacement
    /*----
        case 0x38    : retval = '@';  break;  // steelwall
        case 0x39    : retval = 'S';  break;  // slime
        case 0x0B    : retval = '<';  break;  // ?? butterfly       // should be handled before
        case 0x0A    : retval = '>';  break;  // ?? firefly rechts  // should be handled before
        case 0x33    : retval = '3';  break;  // ??                 // should be handled before
        case 0x32    : retval = '2';  break;  // ??                 // should be handled before
    ----*/
    switch (ch) {
        case 0x38  : el = EL_STEELWALL;                 break;      // defined in "bd_element_mapFull"      // steelwall                    // (rockford?)
        case 0x39  : el = EL_AMOEBA_WET;                break;      // defined in "bd_element_mapFull"      // slime                        // (rockford? scanned this frame)
        // case 0x2A  : el = EL_BD_EXPANDABLE_WALL;        break;      // defined in "bd_element_mapNibble"    // expandWall                   // (undefined?)
        // case 0x3C  : el = EL_PLAYER_1;                  break;      // defined in "bd_element_mapNibble"    // rockford (fake like wall)    // (rockford? fake like wall)
        // case 0x3D  : el = EL_STEEL_EXIT_CLOSED;         break;      // defined in "bd_element_mapNibble"    // hiddenOut                    // (undefined?)
        // case 0x3E  : el = EL_AMOEBA_WET;                break;      // defined in "bd_element_mapNibble"    // slime                        // (undefined?)
        default    : el = elDefault;                    break;
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static int bd_element_mapExtend(int ch, int x, int y, int lev, int elDefault) {
    int el;
    if (elDefault < 0) {
        elDefault = EL_STEEL_CHAR_QUESTION;
    }
    // determine from caves // replacement
    /*----
        case 0x57    : retval = 'A';  break;  // amoeba
    ----*/
    switch (ch) {
    case 0x38    : el = EL_STEELWALL;           break;      // retval = '@';  break;  // steelwall
    case 0x39    : el = EL_AMOEBA_WET;          break;      // retval = 'S';  break;  // slime
    case 0x57    : el = EL_BD_AMOEBA;           break;      // retval = 'A';  break;  // amoeba

    case 0x0B    : el = EL_BD_FIREFLY_4;        break;      // retval = '<';  break;  // ?? butterfly
    case 0x0A    : el = EL_BD_FIREFLY_3;        break;      // retval = '>';  break;  // ?? firefly rechts
    case 0x33    : el = EL_BD_BUTTERFLY_4;      break;      // retval = '3';  break;  // ??

    case 0x56    : el = EL_CHAR_M;              break;      // retval = 'M';  break;  // mrs.rockford
    case 0x6D    : el = EL_CHAR_K;              break;      // retval = 'K';  break;  // key 
    case 0x6E    : el = EL_CHAR_O;              break;      // retval = 'O';  break;  // door 
    case 0x78    : el = EL_CHAR_J;              break;      // retval = 'J';  break;  // jamjam

    case 0x6C    : el = EL_CHAR_D;              break;      // retval = 'D';  break;  // diamond blocked not eatable
    case 0x6B    : el = EL_CHAR_B;              break;      // retval = 'B';  break;  // rock blocked
    case 0x64    : el = EL_CHAR_G;              break;      // retval = 'G';  break;  // pacman ghost
    case 0x3B    : el = EL_CHAR_N;              break;      // retval = 'N';  break;  // bonbon
    case 0x40    : el = EL_CHAR_C;              break;      // retval = 'C';  break;  // classes (water bubbles)

    case 0x32    : el = EL_BD_BUTTERFLY_3;      break;      // retval = '2';  break;  // ?? butterfly
    case 0x29    : el = EL_CHAR_Q;              break;      // retval = 'Q';  break;  // ?? mouse
    case 0x2E    : el = EL_CHAR_E;              break;      // retval = 'h';  break;  // ?? hidden
    case 0x1E    : el = EL_CHAR_T;              break;      // retval = 'T';  break;  // ?? time
    case 0x06    : el = EL_CHAR_6;              break;      // retval = '6';  break;  // ?? 

    default      : el = elDefault;              break;
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static boolean elementUnknown (int el)
{
    boolean retval = FALSE;
    if (el == EL_STEEL_CHAR_QUESTION) {
        retval = TRUE;
    }
    if (el >= EL_STEEL_CHAR_START && el <= EL_STEEL_CHAR_END) {
        retval = TRUE;
    }
    if (el >= EL_CHAR_START && el <= EL_CHAR_END) {
        retval = TRUE;
    }
    return retval;
}

/*----------------------------------------------------------------------------------*/
static int elementFromBCDFF (int ch, int x, int y, int lev, int elDefault, char *ext)
{
    char elementText[20];
    int  el = elDefault;
    sprintf(elementText, "BD%s_0X%2.2X", ext, ch);
    el = BDCFF_itemFromName(elementText, elDefault);
    return el;
}

/*----------------------------------------------------------------------------------*/
/* use "bd_element_" functions - for better debugging                               */
/*----------------------------------------------------------------------------------*/
static int bd_wlc_map(int ch, int x, int y, int lev) {
    int el = -1;
    el = bd_element_mapNibble(ch, x, y, lev, el);  
    if (elementUnknown(el)) {
        el = bd_element_mapFull(ch, x, y, lev, el);
    }
    el = bd_element_mapReplacement(ch, x, y, lev, el);      // Replacement
    if (elementUnknown(el)) {
        el = elementFromBCDFF(ch, x, y, lev, el, "WLC");
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static int bd_wl2_map(int ch, int x, int y, int lev) {
    int el = -1;
    el = bd_element_mapNibble(ch, x, y, lev, el);  
    if (elementUnknown(el)) {
        el = bd_element_mapFull(ch, x, y, lev, el);
    }
    el = bd_element_mapReplacement(ch, x, y, lev, el);      // Replacement
    if (elementUnknown(el)) {
        el = elementFromBCDFF(ch, x, y, lev, el, "WL2");
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static int bd_oto_map(int ch, int x, int y, int lev) {
    int el = -1;
    el = bd_element_mapNibble(ch, x, y, lev, el);  
    if (elementUnknown(el)) {
        el = bd_element_mapFull(ch, x, y, lev, el);
    }
    el = bd_element_mapReplacement(ch, x, y, lev, el);      // Replacement
    if (elementUnknown(el)) {
        el = bd_element_mapExtend(ch, x, y, lev, el);       // OTO -> extend 
    }
    if (elementUnknown(el)) {
        el = elementFromBCDFF(ch, x, y, lev, el, "OTO");
    }
    return el;
}

/*----------------------------------------------------------------------------------*/
static char * toName (char *name) {
  static char cname[12+1]= {0,0,0,0,0,0, 0,0,0,0,0,0, 0};
  int i;

  for (i=0; i < 12; i++) {
    if (*(name+i) > 0) {
      cname[i] = *(name+i);
    } else {
      cname[i] = 0;
      break;
    }
  }
  return cname;
}


/*--- atari -----------------------------------------------*/
static void Load_BDlevel_atari (struct LevelInfo        *level,
                                struct LevelFileInfo    *level_file_info,
                                bd_atari_data_t         *bddata)
{
  int x,y;
  int maxx,maxy;
  int ch,chh,chl;
  int rc = 0;
  boolean intermission;

  maxx = BD_INTERMISSION_WIDTH;
  maxy = BD_INTERMISSION_HEIGHT;
  maxx = BD_CAVE_WIDTH;
  maxy = BD_CAVE_HEIGHT;

  { // test
    char ch = 0;
    int  bcd = 0;
    bcd = bcd1int(ch);
    bcd = 0;
    bcd = color_to_int(ch,bcd);
  }

  // printf("flags:         %3d\n", chint(bddata.flags));
  // printf("name :         %s\n", toName(bddata.cavename));     /*#ATARI*/
  strcpy(level->name, toName(bddata->cavename));

  /*--------------------------------------------------------*/
  /*--- intermission check ---------------------------------*/
  /*--------------------------------------------------------*/
  /*- check1 -*/
  intermission = FALSE;
  if (memcmp(&bddata->flags, bd_atari_flagid_intermission, 1) == 0) {
    // printf("--- Intermission --(1)-\n");
    intermission = TRUE;
  } else {
    // printf("------ Cave -------(1)-\n");
    intermission = FALSE;
  }

  /*- check2 -*/
  intermission = TRUE;
  for (y=0; y<maxy; y++) {
    for (x=BD_INTERMISSION_WIDTH/2; x<maxx/2; x++) {
      ch  = chint (bddata->ldata[y*maxx/2+x]);
      if (ch != 0x44) {
        intermission = FALSE;
        y = maxy+1;
        break;
      }      
    }
  }
  if (intermission) {
    // printf("--- Intermission --(2)-\n");
  } else {
    // printf("------ Cave -------(2)-\n");
  }

  /*---------------------------------------------------------*/
  /*--- output ----------------------------------------------*/
  /*---------------------------------------------------------*/
  if (intermission) {
    level->fieldx = BD_INTERMISSION_WIDTH;
    level->fieldy = BD_INTERMISSION_HEIGHT+1;
  } else {
    level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
    level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy+1), MAX_LEV_FIELDY);
  }

  for (y=0; y<maxy; y++) {
    for (x=0; x<maxx/2; x++) {
      ch  = chint (bddata->ldata[y*maxx/2+x]);
      chh = ch / 16;
      chl = ch % 16;
      // printf("%c%c", atari_map[chh], atari_map[chl]);
      level->field[x*2]  [y] = bd_atari_map(chh);
      level->field[x*2+1][y] = bd_atari_map(chl);
      
      rc += 2;
    }
    // printf("\n");
  }
  /* default (y=maxy) */
  for (x=0; x<maxx/2; x++) {
    // printf("%c%c", atari_map[4], atari_map[4]);   /* wall */
    level->field[x*2]  [y] = bd_atari_map(4);
    level->field[x*2+1][y] = bd_atari_map(4);
  }
  // printf("\n");

  /*---------------------------------------------*/   
  // printf("\n");
  // printf("\n");
  // printf("...read:      %d\n", rc);
  // printf("cavedelay   (00-14) (  0- 20) (00 = fast, 0x14 = slow) :  %3d\n", bcd2int(bddata.cavedelay));
  // printf("cavetime    (00-FF) (  0-255) Cave time in seconds     :  %3d\n", bcd3int(bddata.cavetime));
  level->time                                                                = bcd3int(bddata->cavetime);
  // printf("diamneeded  (00-63) (  0- 99) Diamonds needed          :  %3d\n", bcd2int(bddata.diamneeded));
  level->gems_needed                                                         = bcd2int(bddata->diamneeded);
  // printf("diamvalue   (00-63) (  0- 99) Diamonds value           :  %3d\n", bcd2int(bddata.diamvalue));
  level->score[SC_EMERALD]                                                   = bcd2int(bddata->diamvalue);
  // printf("diambonus   (00-63) (  0- 99) Diamonds bonus           :  %3d\n", bcd2int(bddata.diambonus));
  level->score[SC_TIME_BONUS]                                                = bcd2int(bddata->diambonus);
  // printf("slimperm    (00-08) (  0-  8) Slime permeability       :  %3d\n", chint(bddata.slimperm));
  // printf("amoebatime  (00-FF) (  0-255) Amoeba life time         :  %3d\n", bcd3int(bddata.amoebatime));
  level->amoeba_speed                                                        = bcd3int(bddata->amoebatime);
  level->grow_into_diggable                                                  = TRUE;
  // printf("magwalltime (00-FF) (  0-255) Magic wall milling time  :  %3d\n", bcd3int(bddata.magwalltime));
  level->time_magic_wall                                                     = bcd3int(bddata->magwalltime);
  // printf("flags                                                  :  %3d\n", chint(bddata.flags));
  // printf("color1      (00-0F) (  0- 15) C                        :  %3d\n", color_to_int(bddata.color1, 0xfe));
  // printf("color2      (00-0F) (  0- 15) B                        :  %3d\n", color_to_int(bddata.color2, 0xfe));
  // printf("color3      (00-0F) (  0- 15) D                        :  %3d\n", color_to_int(bddata.color3, 0xfe));
  // printf("color4      (00-0F) (  0- 15) A unused = 0             :  %3d\n", color_to_int(bddata.color4, 0xfe));
  // printf("color5      (00-07) (  0-  7) - unused                 :  %3d\n", color_to_int(bddata.color5, 0xfe));
  // printf("\n");
  // printf("\n");
  
}


/*--- bd64 ------------------------------------------------*/
static void Load_BDlevel_bd64 (struct LevelInfo     *level,
                               struct LevelFileInfo *level_file_info,
                               bd_bd64_data_t       *bddata)
{
  int x,y;
  int maxx,maxy;
  int ch,chh,chl;
  int rc = 0;
  boolean intermission;

  maxx = BD_INTERMISSION_WIDTH;
  maxy = BD_INTERMISSION_HEIGHT;
  maxx = BD_CAVE_WIDTH;
  maxy = BD_CAVE_HEIGHT;

  // printf("flags:         %3d\n", chint(bddata.flags));
  
  /*--------------------------------------------------------*/
  /*--- intermission check ---------------------------------*/
  /*--------------------------------------------------------*/
  /*- check1 -*/
  intermission = FALSE;
  if (memcmp(&bddata->flags, bd_bd64_flagid_intermission, 1) == 0) {
    // printf("--- Intermission --(1)-\n");
    intermission = TRUE;
  } else {
    // printf("------ Cave -------(1)-\n");
    intermission = FALSE;
  }

  /*- check2 -*/
  intermission = TRUE;
  for (y=0; y<maxy; y++) {
    for (x=BD_INTERMISSION_WIDTH/2; x<maxx/2; x++) {
      ch  = chint (bddata->ldata[y*maxx/2+x]);
      if (ch != 0x44) {
        intermission = FALSE;
        y = maxy+1;
        break;
      }      
    }
  }
  if (intermission) {
    // printf("--- Intermission --(2)-\n");
  } else {
    // printf("------ Cave -------(2)-\n");
  }

  /*---------------------------------------------------------*/
  /*--- output ----------------------------------------------*/
  /*---------------------------------------------------------*/
  if (intermission) {
    level->fieldx = BD_INTERMISSION_WIDTH;
    level->fieldy = BD_INTERMISSION_HEIGHT+1;
  } else {
    level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
    level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy+1), MAX_LEV_FIELDY);
  }

  for (y=0; y<maxy; y++) {
    for (x=0; x<maxx/2; x++) {
      ch  = chint (bddata->ldata[y*maxx/2+x]);
      chh = ch / 16;
      chl = ch % 16;
      // printf("%c%c", bd64_map[chh], bd64_map[chl]);
      level->field[x*2]  [y] = bd_bd64_map(chh);
      level->field[x*2+1][y] = bd_bd64_map(chl);

      rc += 2;
    }
    // printf("\n");
  }
  /* default (y=maxy) */
  for (x=0; x<maxx/2; x++) {
    // printf("%c%c", bd64_map[4], bd64_map[4]);   /* wall */
    level->field[x*2]  [y] = bd_bd64_map(4);
    level->field[x*2+1][y] = bd_bd64_map(4);
  }
  // printf("\n");

  /*---------------------------------------------*/ 
  // printf("\n");
  // printf("\n");
  // printf("...read:      %d\n", rc);
  // printf("cavedelay   (00-14) (  0- 20) (00 = fast, 0x14 = slow) :  %3d\n", chint(bddata.cavedelay));
  // printf("cavetime    (00-FF) (  0-255) Cave time in seconds     :  %3d\n", chint(bddata.cavetime));
  level->time                                                                = chint(bddata->cavetime);
  // printf("diamneeded  (00-63) (  0- 99) Diamonds needed          :  %3d\n", chint(bddata.diamneeded));
  level->gems_needed                                                         = chint(bddata->diamneeded);
  // printf("diamvalue   (00-63) (  0- 99) Diamonds value           :  %3d\n", chint(bddata.diamvalue));
  level->score[SC_EMERALD]                                                   = chint(bddata->diamvalue);
  // printf("diambonus   (00-63) (  0- 99) Diamonds bonus           :  %3d\n", chint(bddata.diambonus));
  level->score[SC_TIME_BONUS]                                                = chint(bddata->diambonus);
  // printf("slimperm    (00-08) (  0-  8) Slime permeability       :  %3d\n", chint(bddata.slimperm));
  // printf("amoebatime  (00-FF) (  0-255) Amoeba life time         :  %3d\n", chint(bddata.amoebatime));
  level->amoeba_speed                                                        = chint(bddata->amoebatime);
  level->grow_into_diggable                                                  = TRUE;
  // printf("magwalltime (00-FF) (  0-255) Magic wall milling time  :  %3d\n", chint(bddata.magwalltime));
  level->time_magic_wall                                                     = chint(bddata->magwalltime);
  // printf("flags                                                  :  %3d\n", chint(bddata.flags));
  // printf("color1      (00-0F) (  0- 15) - unused                 :  %3d\n", color_to_int(bddata.color1, 0xff));
  // printf("color2      (00-0F) (  0- 15) - unused                 :  %3d\n", color_to_int(bddata.color2, 0xff));
  // printf("color3      (00-0F) (  0- 15) B                        :  %3d\n", color_to_int(bddata.color3, 0xff));
  // printf("color4      (00-0F) (  0- 15) A                        :  %3d\n", color_to_int(bddata.color4, 0xff));
  // printf("color5      (00-07) (  0-  7) C                        :  %3d\n", color_to_int(bddata.color5, 0xff));
  // printf("\n");
  // printf("\n");

}


/*--- wlc -------------------------------------------------*/
static void Load_BDlevel_wlc (struct LevelInfo      *level,
                              struct LevelFileInfo  *level_file_info,
                              bd_wlc_data_t         *bddata)
{
  int x,y;
  int maxx,maxy;
  int ch;
  int rc = 0;
  boolean intermission;

  maxx = BD_INTERMISSION_WIDTH;
  maxy = BD_INTERMISSION_HEIGHT;
  maxx = BD_CAVE_WIDTH;
  maxy = BD_CAVE_HEIGHT;

  /*--------------------------------------------------------*/
  /*--- intermission check ---------------------------------*/
  /*--------------------------------------------------------*/
  /*- check1 -*/
  intermission = FALSE;
  // cave Q,R,S,T -- 17,18,29,20 are intermissions

  /*---------------------------------------------------------*/
  /*--- output ----------------------------------------------*/
  /*---------------------------------------------------------*/
  if (intermission) {
    level->fieldx = BD_INTERMISSION_WIDTH;
    level->fieldy = BD_INTERMISSION_HEIGHT+1;
  } else {
    level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
    level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy+1), MAX_LEV_FIELDY);
  }

  for (y=0; y<maxy; y++) {
    for (x=0; x<maxx; x++) {
      ch  = chint (bddata->ldata[y*maxx+x]);
      level->field[x][y] = bd_wlc_map(ch, x, y, level_file_info->nr);
      rc += 1;
    }
  }

  /*---------------------------------------------*/ 
  level->time                                                                = chint(bddata->cavetime[0]);
  level->gems_needed                                                         = chint(bddata->diamneeded[0]);
  level->score[SC_EMERALD]                                                   = chint(bddata->diamvalue);
  level->score[SC_TIME_BONUS]                                                = chint(bddata->diambonus);
  level->amoeba_speed                                                        = chint(bddata->magwalltime);   // WLC have only one value for magwalltime/amoebatime
  level->grow_into_diggable                                                  = TRUE;
  level->time_magic_wall                                                     = chint(bddata->magwalltime);

}


/*--- wl2 -------------------------------------------------*/
static void Load_BDlevel_wl2 (struct LevelInfo      *level,
                              struct LevelFileInfo  *level_file_info,
                              bd_wl2_data_t         *bddata)
{
  int x,y;
  int maxx,maxy;
  int ch;
  int rc = 0;
  boolean intermission;

  maxx = BD_INTERMISSION_WIDTH;
  maxy = BD_INTERMISSION_HEIGHT;
  maxx = BD_CAVE_WIDTH;
  maxy = BD_CAVE_HEIGHT;

  /*--------------------------------------------------------*/
  /*--- intermission check ---------------------------------*/
  /*--------------------------------------------------------*/
  /*- check1 -*/
  intermission = FALSE;
  // cave Q,R,S,T -- 17,18,29,20 are intermissions

  /*---------------------------------------------------------*/
  /*--- output ----------------------------------------------*/
  /*---------------------------------------------------------*/
  if (intermission) {
    level->fieldx = BD_INTERMISSION_WIDTH;
    level->fieldy = BD_INTERMISSION_HEIGHT+1;
  } else {
    level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
    level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy+1), MAX_LEV_FIELDY);
  }

  for (y=0; y<maxy; y++) {
    for (x=0; x<maxx; x++) {
      ch  = chint (bddata->ldata[y*maxx+x]);
      level->field[x][y] = bd_wl2_map(ch, x, y, level_file_info->nr);      // same map as WLC
      rc += 1;
    }
  }

  /*---------------------------------------------*/ 
  level->time                                                                = chint(bddata->cavetime[0]);
  level->gems_needed                                                         = chint(bddata->diamneeded[0]);
  level->score[SC_EMERALD]                                                   = chint(bddata->diamvalue);
  level->score[SC_TIME_BONUS]                                                = chint(bddata->diambonus);
  level->amoeba_speed                                                        = chint(bddata->magwalltime);   // WLC have only one value for magwalltime/amoebatime
  level->grow_into_diggable                                                  = TRUE;
  level->time_magic_wall                                                     = chint(bddata->magwalltime);

}


/*--- oto -------------------------------------------------*/
static void Load_BDlevel_oto (struct LevelInfo      *level,
                              struct LevelFileInfo  *level_file_info,
                              bd_oto_data_t         *bddata)
{
  int x,y;
  int maxx,maxy;
  int ch;
  int rc = 0;
  boolean intermission;

  maxx = BD_INTERMISSION_WIDTH;
  maxy = BD_INTERMISSION_HEIGHT;
  maxx = BD_CAVE_WIDTH;
  maxy = BD_CAVE_HEIGHT;

  /*--------------------------------------------------------*/
  /*--- intermission check ---------------------------------*/
  /*--------------------------------------------------------*/
  /*- check1 -*/
  intermission = FALSE;
  // cave Q,R,S,T -- 17,18,29,20 are intermissions

  /*---------------------------------------------------------*/
  /*--- output ----------------------------------------------*/
  /*---------------------------------------------------------*/
  if (intermission) {
    level->fieldx = BD_INTERMISSION_WIDTH;
    level->fieldy = BD_INTERMISSION_HEIGHT+1;
  } else {
    level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
    level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy+1), MAX_LEV_FIELDY);
  }

  for (y=0; y<maxy; y++) {
    for (x=0; x<maxx; x++) {
      ch  = chint (bddata->ldata[y*maxx+x]);
      level->field[x][y] = bd_oto_map(ch, x, y, level_file_info->nr);      // same map as WLC ?? BilderDash // many new elements
      rc += 1;
    }
  }

  /*---------------------------------------------*/
  // time
  // dias
  // extra bonus
  // bonus
  level->time                                                                = bcd3int(&bddata->para[0]);     // time 
  level->gems_needed                                                         = bcd3int(&bddata->para[3]);     // diamonds
  level->score[SC_EMERALD]                                                   = bcd3int(&bddata->para[9]);     // diamValue
  level->score[SC_TIME_BONUS]                                                = bcd3int(&bddata->para[6]);     // diamBonus
  level->amoeba_speed                                                        = 50;                           // ????unknown????
  level->grow_into_diggable                                                  = TRUE;
  level->time_magic_wall                                                     = 100;                          // ????unknown????

}

/*--------------------------------------------------------------*/
static void adjustTimeValues(struct LevelInfo *level)
{
  int t;

  t = level->time;
  t = t + options.bd_caveTimeOffset + t * options.bd_caveTimeGrade/100.0;
  level->time                     = t;

  t = level->amoeba_speed; 
  t = t + options.bd_amoebaTimeOffset + t * options.bd_amoebaTimeGrade/100.0;
  level->amoeba_speed             = t;

  t = level->time_magic_wall; 
  t = t + options.bd_magicwallTimeOffset + t * options.bd_magicwallTimeGrade/100.0;
  level->time_magic_wall          = t;

}


/*--------------------------------------------------------------*/
void LoadLevelFromFileInfo_BD(struct LevelInfo      *level,
                              struct LevelFileInfo  *level_file_info)
{
  char *filename = level_file_info->filename;
  File *file;
  int  readLen;
  char buf[BD_FILE_MAXLEN+1];
  bd_bd64_data_t    bd64data;
  bd_atari_data_t   ataridata;
  bd_wlc_data_t     wlcdata;
  bd_wl2_data_t     wl2data;
  bd_oto_data_t     otodata;

  // init parameter, map
  BDCFF_init();  

  if (!(file = openFile(filename, MODE_READ)))
  {
    level->no_valid_file = TRUE;

    Error("cannot read level '%s' -- using empty level", filename);

    return;
  }

  readLen = readFile(file, &buf[0],1,BD_FILE_MAXLEN);
  if (readLen >= 4) {
    if (memcmp(&buf[0], bd_atari_fileid_head, 4) == 0) {
      /* atari BD file */
      level->name[0] = 0;
      memcpy(&ataridata, &buf[0], sizeof(bd_atari_data_t));
      Load_BDlevel_atari(level, level_file_info, &ataridata);
      adjustTimeValues(level);
      if (level->name[0] == 0) {
        strcpy(level->name, level_file_info->basename);
      }
    } else if (memcmp(&buf[0], bd_bd64_fileid_head, 2) == 0) {
      /* bd64  BD file */
      level->name[0] = 0;
      memcpy(&bd64data, &buf[0], sizeof(bd_bd64_data_t));
      Load_BDlevel_bd64(level, level_file_info, &bd64data);
      adjustTimeValues(level);
      if (level->name[0] == 0) {
        strcpy(level->name, level_file_info->basename);
      }
    } else if (memcmp(&buf[0], bd_wlc_fileid_head, 5) == 0) {
      /* wlc  BD file */
      level->name[0] = 0;
      memcpy(&wlcdata, &buf[0], sizeof(bd_wlc_data_t));
      Load_BDlevel_wlc(level, level_file_info, &wlcdata);
      adjustTimeValues(level);
      if (level->name[0] == 0) {
        strcpy(level->name, level_file_info->basename);
      }
    } else if (memcmp(&buf[0], bd_wl2_fileid_head, 5) == 0) {
      /* wl2  BD file */
      level->name[0] = 0;
      memcpy(&wl2data, &buf[0], sizeof(bd_wl2_data_t));
      Load_BDlevel_wl2(level, level_file_info, &wl2data);
      adjustTimeValues(level);
      if (level->name[0] == 0) {
        strcpy(level->name, level_file_info->basename);
      }
    } else if (memcmp(&buf[0], bd_oto_fileid_head, 5) == 0) {
      /* oto  BD file */
      level->name[0] = 0;
      memcpy(&otodata, &buf[0], sizeof(bd_oto_data_t));
      Load_BDlevel_oto(level, level_file_info, &otodata);
      adjustTimeValues(level);
      if (level->name[0] == 0) {
        strcpy(level->name, level_file_info->basename);
      }
    } else {
      // close to reopen again
      closeFile(file);
      file = NULL;
      buf[BD_FILE_MAXLEN-1] = 0;  // terminate for "strstr"

      if (strstr(buf,"[BDCFF]") || strstr(buf,"[bdcff]")) {
        if (!Load_BDCFFlevel(level, level_file_info)) {
          level->no_valid_file = TRUE;

          Error("cannot read level from file '%s' -- using empty level",
	        filename);

          return;
        }
      } else 
      if (strstr(buf,"GAMEBD") || strstr(buf,"gamebd")) {
        if (!Load_GAMEBDlevel(level, level_file_info)) {
          level->no_valid_file = TRUE;

          Error("cannot read level from file '%s' -- using empty level",
	        filename);

          return;
        }
      } else 
      {
        level->no_valid_file = TRUE;

        Error("cannot read level from file '%s' -- using empty level",
	      filename);

        return;
      }
    }
  } else {
    level->no_valid_file = TRUE;

    Error("cannot read level from file '%s' -- using empty level",
	  filename);

    closeFile(file);
    return;
  }

  if (file) {
    closeFile(file);
  }
}
