// ============================================================================
// filesBDCFF.c
// ----------------------------------------------------------------------------
// (c) 2006-2023 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// loader for BDCFF game files
// ----------------------------------------------------------------------------
// as defined at http://www.emeraldmines.net/
// discused here https://www.artsoft.org/forum/viewtopic.php?t=1547
// ----------------------------------------------------------------------------
// credits to Arno Weber - who help a lot
// ============================================================================

#include <ctype.h>
#include <sys/stat.h>
#include <math.h>

#include "libgame/libgame.h"

#include "files.h"
#include "filesBD.h"
#include "filesBDCFF.h"
#include "init.h"
#include "tools.h"
#include "tape.h"
#include "libgame/rndapi.h"		/*#HAG#VC#*/
#include "libgame/util.h"


#define BDCFF_CAVE_WIDTH            40
#define BDCFF_CAVE_HEIGHT           22
#define BDCFF_INTERMISSION_WIDTH    20
#define BDCFF_INTERMISSION_HEIGHT   12

// Constants
#define MAX_N_LEVELS  5
#define MAX_N_COLORS  5
#define MAX_N_PROBS  10

// Typedef
#define MAX_STRING_LEN  256
#define MAX_ITEM_LEN    40

typedef char string_t[MAX_STRING_LEN];
typedef char item_t[MAX_ITEM_LEN];


// Game properties
typedef struct game_s { 
  string_t  game_name;
  int       count_caves;
  int       n_caves;
  int       n_levels;
  string_t  date;
  string_t  author;
  string_t  www;
  string_t  charset;
  string_t  fontset;
  string_t  remark_game;
  string_t  description_game;
  int       mapCodeLen;
} game_t;


// Cave properties
typedef struct cave_s {
  string_t  cave_name;
  boolean   intermission;
  boolean   intermissionChecked;
  int       delay;
  int       cave_time[MAX_N_LEVELS];
  int       diamonds_required[MAX_N_LEVELS];
  int       diamond_value;
  int       bonus_diamond_value;
  string_t  slime_permeability; // is in fact a float number
  int       amoeba_time;
  int       magic_wall_time;
  string_t  remark_cave;
  string_t  description_cave;
  int       map_matrix[MAX_N_LEVELS][BDCFF_CAVE_HEIGHT][BDCFF_CAVE_WIDTH];
  long      color[MAX_N_COLORS];
  boolean   amoeba_present;
  boolean   slime_present;

  // control vars for cave
  boolean   map_available;
  item_t    initial_fill;
  boolean   use_random_fill;

  // random generator 
  int       rand_seed[MAX_N_LEVELS];
  int       random_fill_items[MAX_N_PROBS];
  int       random_fill_probs[MAX_N_PROBS];
} cave_t;


// map list record
typedef struct mapCodeRec_s {
  int       id;
  item_t    item;
  char      code1[2];
  char      code2[3];
  item_t    sourceEngine;
  int       rndid;
  item_t    elementDescr;
  int       element;
} mapCodeRec_t;


// global variables
static char    bdcff_inputfile[MAX_FILENAME_LEN] = "";
static boolean use_default_colors  = FALSE;   // FALSE = read from file

static game_t  m_game;
static cave_t  m_cave; 
static int     m_numberCaves  = 0;
static int     m_numberLevels = 0;
static int     m_cave_nr;
static int     m_lev;

static int     m_cavx    = BDCFF_CAVE_WIDTH;
static int     m_cavy    = BDCFF_CAVE_HEIGHT;
static int     m_intx    = BDCFF_INTERMISSION_WIDTH;
static int     m_inty    = BDCFF_INTERMISSION_HEIGHT;

static List_t  m_mapList = {0,0,0,0};


// default element for unknown, empty 
#define BDCFF_UNKNOWN           EL_STEEL_CHAR_QUESTION

static int     m_element_space        = EL_EMPTY;


// default color set
static item_t  default_color_set[MAX_N_COLORS] = {"Black","Black","Brown","LightRed","White"};


// State variables
static int   error                 = 0;
static int   warning               = 0;


// define error's
#define ERR_VALUE_EXPECTED           41

#define ERR_VAL_LINE                 33
#define ERR_VAL_FILLRECT             32
#define ERR_VAL_RECTANGLE            31

#define ERR_EOF_IN_CAVE              23
#define ERR_EOF_OR_CAVE              22
#define ERR_EOF_BEFORE_CAVE          21

#define ERR_NOEND_OBJECTS            14
#define ERR_NOEND_MAP                13
#define ERR_NOEND_CAVE               12
#define ERR_NOEND_BDCFF              11

#define ERR_NO_SECTION                7
#define ERR_NO_GAME                   6
#define ERR_NO_BDCFF                  5
#define ERR_NO_CAVES                  4
#define ERR_NO_GAMENAME               3
#define ERR_NO_COLORNAMES             2

#define ERR_FILE_NOTFOUND             1



// define's for read function's (set error)
#define C_EMPTY_TAB       " \t"

#define read_int(s)     rtryInt(s, C_EMPTY_TAB, FALSE, ERR_VALUE_EXPECTED) 
#define read_dbl(s)     rtryDbl(s, C_EMPTY_TAB, FALSE, ERR_VALUE_EXPECTED) 
#define read_str(s)     rtryStr(s, C_EMPTY_TAB, FALSE, ERR_VALUE_EXPECTED) 

#define rtry_int(s)     rtryInt(s, C_EMPTY_TAB, FALSE, 0) 
#define rtry_dbl(s)     rtryDbl(s, C_EMPTY_TAB, FALSE, 0) 
#define rtry_str(s)     rtryStr(s, C_EMPTY_TAB, FALSE, 0) 

#define rdsemi_int(s)   rtryInt(s, ";", TRUE, 0) 
#define rdsemi_dbl(s)   rtryInt(s, ";", TRUE, 0) 
#define rdsemi_str(s)   rtryStr(s, ";", TRUE, 0) 


/*---------------------------------------------------------------------*/
static int rtryInt(char *s, char *del, int nochk, int err)
{  
   char *tok = strtok(s, del);
   if ((tok!=NULL) && (nochk || !strEmpty(tok))) {
     return atoi(tok);
   } else {
     if (err) {
       error = err;
     }
     return 0;
   }
}


static double rtryDbl(char *s, char *del, int nochk, int err)
{  
   char *tok = strtok(s, del);
   if ((tok!=NULL) && (nochk || !strEmpty(tok))) {
     return atof(tok);
   } else {
     if (err) {
       error = err;
     }
     return 0.0;
   }
}


static char* rtryStr(char *s, char *del, int nochk, int err)
{  
   char *tok = strtok(s, del);
   if ((tok != NULL) && (nochk || !strEmpty(tok))) {
     return tok;
   } else {
     if (err) {
       error = err;
     }
     return "";
   }
}



/*----------------------------------------------------------------------*/
static void putItem(int cave_nr, int level_nr, int vert_pos, int hor_pos, int item)
{  
  m_cave.map_matrix[level_nr][vert_pos][hor_pos] = item;
}


static void fill(int cave_nr, int level_nr, int item, int a, int b, int c, int d)
{  
   int i,j;
   for (i=a; i<=b; i++)
     for (j=c; j<=d; j++) 
       putItem(cave_nr, level_nr, i, j, item);
}


/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
static boolean comp1(const void *searchData,
                     const void *listData)
{ /* for mapCode1 */
  mapCodeRec_t *m = (mapCodeRec_t*) listData;
  if (strcmp(m->code1,searchData) == 0)   // case sensitive!!
    return TRUE;
  return FALSE;
} // comp1

static boolean comp2(const void *searchData,
                     const void *listData)
{ /* for mapCode2 */
  mapCodeRec_t *m = (mapCodeRec_t*) listData;
  if (strcmp(m->code2,searchData) == 0)   // case sensitive!!
    return TRUE;
  return FALSE;
} // comp2

static boolean compItem(const void *searchData,
                        const void *listData)
{ /* for mapCodeItem */
  mapCodeRec_t *m = (mapCodeRec_t*) listData;
  if (UTILstrCaseCmp(m->item,searchData) == 0)  // not case sensitive
    return TRUE;
  return FALSE;
} // compItem


/*----------------------------------------------------------------------*/
static int item_from_name(char *str)
{  
  mapCodeRec_t  *mapCodeRec = NULL;

  if (listSearch(&m_mapList, str, TRUE, TRUE, compItem)) {
    if (listGet(&m_mapList, (void**) &mapCodeRec)) {
      return (mapCodeRec->element);
    } 
  } 
  if (UTILstrCaseCmp(str, "UNKNOWN") == 0) {
    return (BDCFF_UNKNOWN);
  } else {
    char *value = NULL;
    strLower(str);
    value = getHashEntry(element_token_hash, str);
    if (value != NULL) {
      return (atoi(value));
    }

    return (item_from_name("UNKNOWN"));
  }
}


/*----------------------------------------------------------------------*/
static int determine_item(char *str, int *j)
{ 
  char          mapCode[3]  = {0,0,0};
  mapCodeRec_t  *mapCodeRec = NULL;

  if (m_game.mapCodeLen == 2) {
    mapCode[0] = str[*j]; *j += 1;
    mapCode[1] = str[*j]; *j += 1;

    if (strcmp(mapCode, "  ") == 0) {
      return m_element_space;
    }
    if (listSearch(&m_mapList, mapCode, TRUE, TRUE, comp2)) {
      if (listGet(&m_mapList, (void**) &mapCodeRec)) {
        return (mapCodeRec->element);
      } 
    } 
  } else {
    mapCode[0] = str[*j]; *j += 1;
    mapCode[1] = 0;

    if (strcmp(mapCode, " ") == 0) {
      return m_element_space;
    }
    if (listSearch(&m_mapList, mapCode, TRUE, TRUE, comp1)) {
      if (listGet(&m_mapList, (void**) &mapCodeRec)) {
        return (mapCodeRec->element);
      } 
    } 
  }
  return (item_from_name("UNKNOWN"));
}




/*----------------------------------------------------------------------*/
static void NextRandom(int *RandSeed1, int *RandSeed2) // Random generator used on C64
{  
   short TempRand1;
   short TempRand2;
   short carry;
   short result;
   if(((*RandSeed1 < 0x00) || (*RandSeed1 > 0xFF))) *RandSeed1 &= 0xFF;
   if(((*RandSeed2 < 0x00) || (*RandSeed2 > 0xFF))) *RandSeed2 &= 0xFF;
   /// assert(((*RandSeed1 >= 0x00) && (*RandSeed1 <= 0xFF)));
   /// assert(((*RandSeed2 >= 0x00) && (*RandSeed2 <= 0xFF)));
   TempRand1 = (*RandSeed1 & 0x0001) * 0x0080;
   TempRand2 = (*RandSeed2 >> 1) & 0x007F;
   result = (*RandSeed2) + (*RandSeed2 & 0x0001) * 0x0080;
   carry = (result > 0x00FF);
   result = result & 0x00FF;
   result = result + carry + 0x13;
   carry = (result > 0x00FF);
   *RandSeed2 = result & 0x00FF;
   result = *RandSeed1 + carry + TempRand1;
   carry = (result > 0x00FF);
   result = result & 0x00FF;
   result = result + carry + TempRand2;
   *RandSeed1 = result & 0x00FF;
   /// assert(((*RandSeed1 >= 0x00) && (*RandSeed1 <= 0xFF)));
   /// assert(((*RandSeed2 >= 0x00) && (*RandSeed2 <= 0xFF)));
   if(((*RandSeed1 < 0x00) || (*RandSeed1 > 0xFF))) *RandSeed1 &= 0xFF;
   if(((*RandSeed2 < 0x00) || (*RandSeed2 > 0xFF))) *RandSeed2 &= 0xFF;
}

static void random_fill(int cave_nr, int level_nr)
{  
   int      RandSeed1;
   int      RandSeed2;
   int      item, itemSteelwall, itemDirt;
   int      number_rand_items=0;
   int      k,j,i;
   cave_t  *cave = &m_cave;

   for (k=0; k<4; k++)
   {  
     if (cave->random_fill_probs[k] <= 255) number_rand_items++;
   }
   RandSeed1 = 0;
   RandSeed2 = cave->rand_seed[level_nr];

   itemSteelwall = item_from_name("STEELWALL");
   itemDirt      = item_from_name("DIRT");

   for (j=0; j<m_cavx; j++)
      putItem(cave_nr, level_nr, 0, j, itemSteelwall);

   for (i=1; i<m_cavy; i++)
   {  for (j=0; j<m_cavx; j++)
      {  
         item = itemDirt; 
         NextRandom(&RandSeed1, &RandSeed2);
         for (k=0; k<number_rand_items; k++)
         {  
            if (RandSeed1 < cave->random_fill_probs[k])
            {  
              item = cave->random_fill_items[k];
            }
         }
         if ( (!cave->intermission && i>=1 && i<=(m_cavy-2) && j>=1 && j<=(m_cavx-2)) ||
              ( cave->intermission && i>=1 && i<=(m_inty-2) && j>=1 && j<=(m_intx-2)) )
            putItem(cave_nr, level_nr, i, j, item);
         else 
            putItem(cave_nr, level_nr, i, j, itemSteelwall);
      }
   }
}


/*----------------------------------------------------------------------*/
static long determine_color(char* s)
{  
   if (!UTILstrCaseCmp(s,"black"     )) return 0x000000;
   if (!UTILstrCaseCmp(s,"white"     )) return 0xFFFFFF;
   if (!UTILstrCaseCmp(s,"red"       )) return 0x9B4D42;
   if (!UTILstrCaseCmp(s,"cyan"      )) return 0x8CD6DF;
   if (!UTILstrCaseCmp(s,"purple"    )) return 0x9D54C4;
   if (!UTILstrCaseCmp(s,"green"     )) return 0x78BF4E;
   if (!UTILstrCaseCmp(s,"blue"      )) return 0x4B3DB7;
   if (!UTILstrCaseCmp(s,"yellow"    )) return 0xE9F582;
   if (!UTILstrCaseCmp(s,"orange"    )) return 0xA36E30;
   if (!UTILstrCaseCmp(s,"brown"     )) return 0x6B5500;
   if (!UTILstrCaseCmp(s,"lightred"  )) return 0xD2887E;
   if (!UTILstrCaseCmp(s,"gray1"     )) return 0x646464;
   if (!UTILstrCaseCmp(s,"gray2"     )) return 0x929292;
   if (!UTILstrCaseCmp(s,"lightgreen")) return 0xC1FF9A;
   if (!UTILstrCaseCmp(s,"lightblue" )) return 0x8D80F3;
   if (!UTILstrCaseCmp(s,"gray3"     )) return 0xC1C1C1;
   else
   {  
     warning=1; 
     return 0x929292;
   }
}

static void initialize(int cave_nr)
{  
   cave_t  *cave = &m_cave;
   int      m,i;

   strcpy(cave->cave_name,"");
   cave->intermission         = FALSE;
   cave->intermissionChecked  = FALSE;
   cave->delay                = -1;
   for (m=0; m<MAX_N_LEVELS; m++) cave->cave_time[m] = -1;
   for (m=0; m<MAX_N_LEVELS; m++) cave->diamonds_required[m] = -1;
   cave->diamond_value        = -1;
   cave->bonus_diamond_value  = -1;
   strcpy(cave->slime_permeability,"");
   cave->amoeba_time          = -1;
   cave->magic_wall_time      = -1;
   strcpy(cave->remark_cave,"");
   strcpy(cave->description_cave,"");
   for (i=0; i<MAX_N_COLORS; i++) {
      cave->color[i] = determine_color(default_color_set[i]);
   }
   cave->amoeba_present       = FALSE;
   cave->slime_present        = FALSE;

   // memset(&cave->map_matrix, 0, sizeof(cave->map_matrix));

   cave->map_available        = FALSE;
   strcpy(cave->initial_fill,"DIRT");
   cave->use_random_fill      = FALSE;

   for (m=0; m<MAX_N_LEVELS; m++) cave->rand_seed[m] = 0;
   for (i=0; i<MAX_N_PROBS; i++)
   {  
      cave->random_fill_items[i] = 0;
      cave->random_fill_probs[i] = 300;
   }

}


/*----------------------------------------------------------------------*/
static void process_map(File *file, int cave_nr, int n_levels)
{  
   string_t s;
   int      y, x;
   int      m,pos;
   int      item;

   y=0;
   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if (!UTILstrCaseCmp(s,"[/map]")) break;
      if (checkEndOfFile(file)) error=ERR_NOEND_MAP;

      pos=0;
      x=0;
      while (pos < strlen(s))
      {  
         if (y<m_cavy && x<m_cavx)
         {  
            item = determine_item(s, &pos);
            for (m=0; m<n_levels; m++)
              putItem(cave_nr, m, y, x, item);
            x++;
         }
      }
      if (x>m_cavx) warning=1;
      if (!strEmpty(s)) y++;
      if (y>m_cavy) warning=2;
   }
}


/*----------------------------------------------------------------------*/
static void process_objects(File *file, int cave_nr, int n_levels)
{  
   cave_t   *cave = &m_cave;
   string_t s,u,v;
   char     *ps, *pt;
   int      i,incx,incy,a,b,c,d,e,f;
   int      lev,n,m,j;
   int      itemu, itemv;

   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if (!UTILstrCaseCmp(s,"[/objects]")) break;
      if (checkEndOfFile(file)) error=ERR_NOEND_OBJECTS;

      if ((ps=strchr(s,'='))!=NULL) 
      {
        *ps = 0;
        strTrim(s);
        pt = ps+1;
        strCut(pt, " /t");

        if (!UTILstrCaseCmp(s,"point"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           itemu = item_from_name(read_str(NULL));

           for (lev=0; lev<n_levels; lev++)
              putItem(cave_nr, lev, b, a, itemu);
        }
        if (!UTILstrCaseCmp(s,"line"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           c = read_int(NULL);
           d = read_int(NULL);           
           itemu = item_from_name(read_str(NULL));

           incx  = (a==c ? 0 : (c>a ? 1 : -1));
           incy  = (b==d ? 0 : (d>b ? 1 : -1));
           if (a!=c && b!=d && abs(a-c)!=abs(b-d)) error=ERR_VAL_LINE;

           for (n=0; n<=abs(c-a); n++)
           {  for (m=0; m<=abs(d-b); m++)
              {  if ( n==m || (a==c || b==d))
                    for (lev=0; lev<n_levels; lev++)
                       putItem(cave_nr, lev, b+m*incy, a+n*incx, itemu);
              }
           }
        }
        if (!UTILstrCaseCmp(s,"rectangle"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           c = read_int(NULL);
           d = read_int(NULL);
           itemu = item_from_name(read_str(NULL));

           if (a<=c && b<=d)
           {  for (n=a; n<=c; n++)
              {  for (m=b; m<=d; m++)
                 {  if (n==a || m==b || n==c || m==d)
                       for (lev=0; lev<n_levels; lev++)
                          putItem(cave_nr, lev, m, n, itemu);
                 }
              }
           }
           else 
             error=ERR_VAL_RECTANGLE;
        }
        if (!UTILstrCaseCmp(s,"fillrect"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           c = read_int(NULL);
           d = read_int(NULL);
           strcpy (u, read_str(NULL));
           itemu = item_from_name(u);
           strcpy (v, rtry_str(NULL)); 
           if (!strEmpty(v)) {
             itemv = item_from_name(v);
           } else {
             itemv = itemu;
           }

           if (a<=c && b<=d)
           {  for (n=a; n<=c; n++)
              {  for (m=b; m<=d; m++)
                 {  if (n==a || m==b || n==c || m==d)
                       for (lev=0; lev<n_levels; lev++)
                          putItem(cave_nr, lev, m, n, itemu);
                    else 
                       for (lev=0; lev<n_levels; lev++)
                          putItem(cave_nr, lev, m, n, itemv);
                 }
              }
           }
           else 
             error=ERR_VAL_FILLRECT;
        }
        if (!UTILstrCaseCmp(s,"raster"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           c = read_int(NULL);
           d = read_int(NULL);
           e = read_int(NULL);
           f = read_int(NULL);
           itemu = item_from_name(read_str(NULL));

           for (n=0; n<c; n++)
           {  for (m=0; m<d; m++)
                 for (lev=0; lev<n_levels; lev++)
                    putItem(cave_nr, lev, (b+m*f), (a+n*e), itemu);
           }
        }
        if (!UTILstrCaseCmp(s,"add"))
        {  
           a = read_int(pt);
           b = read_int(NULL);
           itemu = item_from_name(read_str(NULL));
           itemv = item_from_name(read_str(NULL));

           for (lev=0; lev<n_levels; lev++)
           {  for (i=1; i<m_cavy; i++)
              {  for (j=1; j<m_cavx; j++)
                 {  if ( (cave->map_matrix[lev][i][j] == itemu) && (
                         (!cave->intermission && i+b >= 1 && i+b <= (m_cavy-2)
                                              && j+a >= 1 && j+a <= (m_cavx-2)) ||
                         ( cave->intermission && i+b >= 1 && i+b <= (m_inty-2)
                                              && j+a >= 1 && j+a <= (m_intx-2)) ) )
                       putItem(cave_nr, lev, i+b, j+a, itemv);
                 }
              }
           }
        }
      }  // keyword
   }
}


/*----------------------------------------------------------------------*/
static void read_game_properties(File *file, game_t *game)
{  
   string_t s;
   char     *ps, *pt;

   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);

      if ((ps=strchr(s,'='))!=NULL) 
      {
        *ps = 0;
        strTrim(s);
        pt = ps+1;
        strCut(pt, " /t");

        if (!UTILstrCaseCmp(s,"name"))
        {  
           strcpy(game->game_name, pt);
           strTrim(game->game_name);
        }
        if (!UTILstrCaseCmp(s,"date"))
           strcpy(game->date, pt);
        if (!UTILstrCaseCmp(s,"author"))
           strcpy(game->author, pt);
        if (!UTILstrCaseCmp(s,"www"))
           strcpy(game->www, pt); 
        if (!UTILstrCaseCmp(s,"charset"))
           strcpy(game->charset, pt);
        if (!UTILstrCaseCmp(s,"fontset"))
           strcpy(game->fontset, pt);
        if (!UTILstrCaseCmp(s,"remark"))
           strcpy(game->remark_game, pt);
        if (!UTILstrCaseCmp(s,"description"))
           strcpy(game->description_game, pt);
        if (!UTILstrCaseCmp(s,"levels"))
        {  
           game->n_levels = read_int(pt);
        }
        if (!UTILstrCaseCmp(s,"caves"))
        {  
           game->n_caves  = read_int(pt);
        }
      }  // keyword

      strTrim(s);
      if (strEmpty(s)) break; // stop at white line (important!!!)
      if (checkEndOfFile(file) || !UTILstrCaseCmp(s,"[cave]")) error=ERR_EOF_OR_CAVE;
   }
   if (strEmpty(game->game_name)) error=ERR_NO_GAMENAME;
   
}


/*----------------------------------------------------------------------*/
static void make_map_matrix_all_levels(File *file, int cave_nr, int n_levels)
{  
   cave_t  *cave = &m_cave;
   int      m,i,j;
   int      itemSteelwall = item_from_name("STEELWALL");
   int      itemAmoeba    = item_from_name("AMOEBA");
   int      itemSlime     = item_from_name("SLIME");

   if (cave->map_available)
   {  
      for (m=0; m<n_levels; m++)
         fill(cave_nr, m, itemSteelwall, 0, m_cavy-1, 0, m_cavx-1);
      process_map(file, cave_nr, n_levels);
   }
   else
   {  
      if (cave->use_random_fill)
         for (m=0; m<n_levels; m++) random_fill(cave_nr, m);
      else
      {  
         for (m=0; m<n_levels; m++)
         {  
            fill(cave_nr, m, itemSteelwall, 0, m_cavy-1, 0, m_cavx-1);
            if (cave->intermission)
               fill(cave_nr, m, item_from_name(cave->initial_fill), 1, m_inty-2, 1, m_intx-2);
            else 
               fill(cave_nr, m, item_from_name(cave->initial_fill), 1, m_cavy-2, 1, m_cavx-2);
         }
      }
      process_objects(file, cave_nr, n_levels);
   }
   for (i=0; i<m_cavy; i++) {  
      for (j=0; j<m_cavx; j++) {  
         if (cave->map_matrix[0][i][j] == itemAmoeba) cave->amoeba_present = TRUE;
         if (cave->map_matrix[0][i][j] == itemSlime ) cave->slime_present  = TRUE;
      }
   }
}


/*----------------------------------------------------------------------*/
static void read_cave_properties(File *file, int cave_nr, int n_levels)
{  
   cave_t   *cave = &m_cave;
   string_t s,u;
   char     *ps, *pt;
   item_t   color_names[MAX_N_COLORS] = {"","","","",""};
   int      number_colors;
   int      m,j;

   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);

      if ((ps=strchr(s,'='))!=NULL) 
      {
        *ps = 0;
        strTrim(s);
        pt = ps+1;
        strCut(pt, " /t");

        if (!UTILstrCaseCmp(s,"name"))
           strcpy(cave->cave_name, pt);
        if (!UTILstrCaseCmp(s,"remark"))
           strcpy(cave->remark_cave, pt);
        if (!UTILstrCaseCmp(s,"description"))
           strcpy(cave->description_cave, pt);
        if (!UTILstrCaseCmp(s,"initialfill"))
           strcpy(cave->initial_fill, pt);
        if (!UTILstrCaseCmp(s,"cavedelay"))
        {  
           cave->delay = read_int(pt);
        }
        if (!UTILstrCaseCmp(s,"cavetime"))
        {  
           cave->cave_time[0] = read_int(pt);
           for (m=1; m<n_levels; m++)
              cave->cave_time[m] = read_int(NULL);
        }
        if (!UTILstrCaseCmp(s,"diamondsrequired"))
        {  
           cave->diamonds_required[0] = read_int(pt);
           for (m=1; m<n_levels; m++)
              cave->diamonds_required[m] = read_int(NULL);
        }
        if (!UTILstrCaseCmp(s,"diamondvalue"))
        {  
           cave->diamond_value        = read_int(pt);
           // Bonus value not given: it will be 0.
           cave->bonus_diamond_value  = rtry_int(NULL);
        }
        if (!UTILstrCaseCmp(s,"slimepermeability"))
           strcpy(cave->slime_permeability, pt);
        if (!UTILstrCaseCmp(s,"amoebatime"))
           cave->amoeba_time     = read_int(pt); 
        if (!UTILstrCaseCmp(s,"magicwalltime"))
           cave->magic_wall_time = read_int(pt);
        if (!UTILstrCaseCmp(s,"intermission"))
        {  
           if (!UTILstrCaseCmp(pt,"true"))
              cave->intermission = TRUE;
           else 
              cave->intermission = FALSE;
        }
        if (!UTILstrCaseCmp(s,"colors"))
        {  
           number_colors=0;
           while (!error)
           {  
              strcpy(color_names[number_colors], rtry_str(pt));
              if (strEmpty(color_names[number_colors])) break;
              number_colors++;
              pt = NULL;
           }
           if (number_colors < MAX_N_COLORS)
           {  
              for (j=0; j < (MAX_N_COLORS-number_colors-1); j++)
              {  
                 if (!use_default_colors)
                    cave->color[j] = determine_color("Black");
              }
           }
           for (j=0; j < number_colors; j++)
           {  
              if (!use_default_colors)
                 cave->color[MAX_N_COLORS-number_colors+j] = determine_color(color_names[j]);
           }
        }
        if (!UTILstrCaseCmp(s,"randseed"))
        {  
           cave->rand_seed[0] = read_int(pt);
           for (m=1; m<n_levels; m++)
              cave->rand_seed[m] = read_int(NULL);
        }
        if (!UTILstrCaseCmp(s,"randomfill"))
        {  
           j=0;
           while(!error)
           {  
              strcpy(u,rtry_str(pt));
              if (!strEmpty(u)) {
                cave->random_fill_items[j] = item_from_name(u);
                cave->random_fill_probs[j] = read_int(NULL);
                j++;
              } else {
                break;
              }
              pt = NULL;
           }
           cave->use_random_fill = TRUE;
        }
        // Later: arrays random items en probs sorteren naar dalende probs.
        // sort array use probs (asc/desc?) 

      }   // keyword

      if (!UTILstrCaseCmp(s,"[map]"))
      {  
         cave->map_available = TRUE;
         break;
      }
      if (!UTILstrCaseCmp(s,"[objects]"))
      {  
         cave->map_available = FALSE;
         break;
      }
      if (checkEndOfFile(file)) error=ERR_EOF_IN_CAVE;
   }
   if (strEmpty(color_names[0])) error=ERR_NO_COLORNAMES;
}


/*----------------------------------------------------------------------*/
static boolean process_cave(File *file, int cave_nr, int n_levels)
{  
   string_t s;
   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if (!UTILstrCaseCmp(s,"[/game]"))
      {  
         if (m_game.count_caves == 0) 
           error=ERR_NO_CAVES;
         else 
           return FALSE;
      }
      if (checkEndOfFile(file)) error=ERR_EOF_BEFORE_CAVE;
      if (!UTILstrCaseCmp(s,"[cave]")) break;
   }

   initialize(cave_nr);
   read_cave_properties(file, cave_nr, n_levels);
   make_map_matrix_all_levels(file, cave_nr, n_levels);

   while (!error)
   {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if (!UTILstrCaseCmp(s,"[/cave]")) break;
      if (checkEndOfFile(file)) error=ERR_NOEND_CAVE;
   }

   return TRUE;
}


/*----------------------------------------------------------------------*/
static void process_game(void)
{  
   File *file;
   string_t s;

   file = openFile(bdcff_inputfile, "r");
   if (file != NULL) {  

     while (!error)
     {  
        getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
        if (!UTILstrCaseCmp(s,"[BDCFF]")) break;
        if (checkEndOfFile(file)) error=ERR_NO_BDCFF;
     }
     while (!error)
     {  
        getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
        if (!UTILstrCaseCmp(s,"[game]")) break;
        if (checkEndOfFile(file)) error=ERR_NO_GAME;
     }

     m_game.count_caves = 0;
     if (!error) {
       read_game_properties(file, &m_game);
       while (!error && process_cave(file,  m_game.count_caves, m_game.n_levels)) {
         if ((m_cave_nr != -1) && (m_cave_nr == m_game.count_caves)) break;
         m_game.count_caves++;
       }
     }

     while (!error)
     {  
        getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
        if (!UTILstrCaseCmp(s,"[/BDCFF]")) break;
        if (checkEndOfFile(file)) error=ERR_NOEND_BDCFF;
     }
     closeFile(file);
   } else {
     error=ERR_FILE_NOTFOUND;
   }
}


/*----------------------------------------------------------------------*/
static void bdcff_mapAdd(int id, char* code1, char* code2, char* item, char *engine, int ele)
{
  mapCodeRec_t  mapCodeRec;

  mapCodeRec.id      = id;
  strcpy(mapCodeRec.code1, code1);
  strcpy(mapCodeRec.code2, code2);
  strcpy(mapCodeRec.item,  item);
  strcpy(mapCodeRec.sourceEngine, engine);
  mapCodeRec.element = ele;

  listEnter(&m_mapList, &mapCodeRec, sizeof(mapCodeRec_t), FALSE);
}

 
static void bdcff_mapInit(void) 
{
  char *NOP = "";
  int  id   = 0;

  if (listEmpty(&m_mapList)) 
  {
    bdcff_mapAdd(id++, NOP , NOP , "UNKNOWN",             "BD", EL_STEEL_CHAR_QUESTION);

    bdcff_mapAdd(id++, "P" , NOP , "INBOX",               "BD", EL_PLAYER_1);
    bdcff_mapAdd(id++, "X" , NOP , "OUTBOX",              "BD", EL_EM_EXIT_CLOSED);
    bdcff_mapAdd(id++, "H" , NOP , "HIDDENOUTBOX",        "BD", EL_EM_STEEL_EXIT_CLOSED);
    bdcff_mapAdd(id++, "W" , NOP , "STEELWALL",           "BD", EL_STEELWALL);
    bdcff_mapAdd(id++, NOP , NOP , "WALL2",               "BD", EL_STEELWALL);
    bdcff_mapAdd(id++, " " , NOP , "SPACE",               "BD", EL_EMPTY);
    bdcff_mapAdd(id++, "." , NOP , "DIRT",                "BD", EL_SAND);
    bdcff_mapAdd(id++, "r" , NOP , "BOULDER",             "BD", EL_BD_ROCK);
    bdcff_mapAdd(id++, NOP , NOP , "BOULDERf",            "BD", EL_BD_ROCK);
    bdcff_mapAdd(id++, NOP , NOP , "FALLING_BOULDER",     "BD", EL_BD_ROCK);
    bdcff_mapAdd(id++, "d" , NOP , "DIAMOND",             "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "DIAMONDf",            "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "FALLING_DIAMOND",     "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "EXPLOSION5D",         "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "EXPLOSION4D",         "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "EXPLOSION3D",         "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "EXPLOSION2D",         "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, NOP , NOP , "EXPLOSION1D",         "BD", EL_BD_DIAMOND);
    bdcff_mapAdd(id++, "w" , NOP , "WALL",                "BD", EL_BD_WALL);
    bdcff_mapAdd(id++, "M" , NOP , "MAGICWALL",           "BD", EL_BD_MAGIC_WALL);
    bdcff_mapAdd(id++, NOP , NOP , "MAGIC_WALL",          "BD", EL_BD_MAGIC_WALL);
    bdcff_mapAdd(id++, "x" , NOP , "HEXPANDINGWALL",      "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, NOP , NOP , "HEXPANDING_WALL",     "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, "v" , NOP , "VEXPANDINGWALL",      "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, NOP , NOP , "VEXPANDING_WALL",     "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, "V" , NOP , "EXPANDINGWALL",       "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, NOP , NOP , "EXPANDING_WALL",      "BD", EL_BD_EXPANDABLE_WALL);
    bdcff_mapAdd(id++, "Q" , NOP , "FIREFLYL",            "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, "O" , NOP , "FIREFLYR",            "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, "o" , NOP , "FIREFLYU",            "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, "q" , NOP , "FIREFLYD",            "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_FIREFLYL",    "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_FIREFLYR",    "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_FIREFLYU",    "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_FIREFLYD",    "BD", EL_BD_FIREFLY);
    bdcff_mapAdd(id++, "C" , NOP , "BUTTERFLYL",          "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, "B" , NOP , "BUTTERFLYR",          "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, "b" , NOP , "BUTTERFLYU",          "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, "c" , NOP , "BUTTERFLYD",          "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_BUTTERFLYL",  "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_BUTTERFLYR",  "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_BUTTERFLYU",  "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, NOP , NOP , "SCANNED_BUTTERFLYD",  "BD", EL_BD_BUTTERFLY);
    bdcff_mapAdd(id++, "a" , NOP , "AMOEBA",              "BD", EL_BD_AMOEBA);
    bdcff_mapAdd(id++, "s" , NOP , "SLIME",               "BD", EL_QUICKSAND_EMPTY);
    bdcff_mapAdd(id++, "F" , NOP , "DUMMY",               "BD", EL_PLAYER_1);

  }

} // bdcff_mapInit


static int bdcff_mapLoad(void) 
{
  char          *filename;
  File          *file;
  mapCodeRec_t  mapCodeRec;
  string_t      s;
  char          *ps, *pt;

  filename = getPath2(options.base_directory, "bdcff.conf");

  file = openFile(filename, "r");
  if (file) {
    while (!error)
    {  
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if (!UTILstrCaseCmp(s,"[parameter]")) break;
      if (checkEndOfFile(file)) error=ERR_NO_SECTION;
    }

    while (!error)
    {
      getStringFromFile(file, s, sizeof(s)); strRemoveNL(s);
      if ((ps=strchr(s,'='))!=NULL) 
      {
        *ps = 0;
        strTrim(s);
        pt = ps+1;
        strCut(pt, " /t");

        if (!UTILstrCaseCmp(s,"caveTime"))
        {
          options.bd_caveTimeOffset = read_int(pt);
          options.bd_caveTimeGrade  = read_int(NULL);  /* devided with 100 */
        }
        if (!UTILstrCaseCmp(s,"amoebaTime"))
        {
          options.bd_amoebaTimeOffset = read_int(pt);
          options.bd_amoebaTimeGrade  = read_int(NULL);  /* devided with 100 */
        }
        if (!UTILstrCaseCmp(s,"magicwallTime"))
        {
          options.bd_magicwallTimeOffset = read_int(pt);
          options.bd_magicwallTimeGrade  = read_int(NULL);  /* devided with 100 */
        }
      }
      if (!UTILstrCaseCmp(s,"[element]")) break;
      if (checkEndOfFile(file)) error=ERR_NO_SECTION;
    }
      
    while (!error && !checkEndOfFile(file)) {
      if (getStringFromFile(file, s, sizeof(s)) != NULL) {
        strRemoveNL(s);
        if (!strEmpty(s)) {
          mapCodeRec.id      = rdsemi_int(s);
          strcpy(mapCodeRec.item,  rdsemi_str(NULL));
          strcpy(mapCodeRec.code1, rdsemi_str(NULL));
          strcpy(mapCodeRec.code2, rdsemi_str(NULL));
          strcpy(mapCodeRec.sourceEngine, rdsemi_str(NULL));
          mapCodeRec.rndid   = rdsemi_int(NULL);
          strcpy(mapCodeRec.elementDescr, rdsemi_str(NULL));

          strTrim(mapCodeRec.item);
          strTrim(mapCodeRec.sourceEngine);
          strTrim(mapCodeRec.elementDescr);
          strUpper(mapCodeRec.sourceEngine);
          strLower(mapCodeRec.elementDescr);

          if (!strEmpty(mapCodeRec.elementDescr)) {
            char *value = NULL;
            value = getHashEntry(element_token_hash, mapCodeRec.elementDescr);
            if (value != NULL) {
              mapCodeRec.element = atoi(value);
            } else {               
              mapCodeRec.element = atoi(mapCodeRec.elementDescr);
              if (mapCodeRec.element == 0) {
                strLower(mapCodeRec.item);
                value = getHashEntry(element_token_hash, mapCodeRec.item);
                strUpper(mapCodeRec.item);
                if (value != NULL) {
                  mapCodeRec.element = atoi(value);
                } else {
                  mapCodeRec.element = BDCFF_UNKNOWN;
                }
              }
            }
          } else {
            mapCodeRec.element = BDCFF_UNKNOWN;
          }

          if (!error) {
            listEnter(&m_mapList, &mapCodeRec, sizeof(mapCodeRec_t), FALSE);
          }
        }
      }
    }
    closeFile(file);
  } else {
    error = ERR_FILE_NOTFOUND;
  }
  checked_free(filename);

  return (error == 0);
}


/*----------------------------------------------------------------------*/
void BDCFF_init(void) 
{
  if (listEmpty(&m_mapList)) {
    mapCodeRec_t  *mapCodeRec;

    if (!bdcff_mapLoad()) {
      bdcff_mapInit();
    }

    /* search for SPACE */
    listHead(&m_mapList);
    if (listSearch(&m_mapList, "SPACE", TRUE, FALSE, compItem)) {
      if (listGet(&m_mapList, (void**) &mapCodeRec)) {
        m_element_space = mapCodeRec->element;
      } 
    } 
  }
}  


/*----------------------------------------------------------------------*/
int BDCFF_itemFromName(char *str, int eledefault)
{
  mapCodeRec_t  *mapCodeRec = NULL;
  char elementStr[MAX_ITEM_LEN+1] = {0};
  strcpy(elementStr, str);

  if (listSearch(&m_mapList, elementStr, TRUE, TRUE, compItem)) {
    if (listGet(&m_mapList, (void**) &mapCodeRec)) {
      return (mapCodeRec->element);
    } 
  } 
  return eledefault;
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
boolean Load_BDCFFlevel(struct LevelInfo *level,
                        struct LevelFileInfo *level_file_info)
{
    int     maxx, maxy, x,y;
    int     itemSteelwall;
    int     t;

    error        = 0;
    warning      = 0;

    /* map init */
    BDCFF_init();

    error        = 0;
    warning      = 0;

    {   // test
        char test[] = "1";
        double d = rtry_dbl(test);
        error = d * 0;
    }

    use_default_colors = FALSE;
    memset(&m_game, 0, sizeof(m_game));

    if (strcmp(bdcff_inputfile, level_file_info->filename) == 0) {
      // ok 
    } else {
      strcpy (bdcff_inputfile, level_file_info->filename);
      m_numberCaves  = 0;
      m_numberLevels = 0;
      m_cave_nr      = -1;
      m_lev          = -1;
      process_game();   // read all!
      m_numberCaves  = m_game.count_caves; 
      m_numberLevels = m_game.n_levels;
    }

    if (m_numberCaves <= 0) {
      error = ERR_NO_CAVES;
    }

    if(error == 0) {
      if (level_file_info->nr <= (m_numberCaves * m_numberLevels)) 
      {
        m_cave_nr = (level_file_info->nr - 1) % m_numberCaves;
        m_lev     = (level_file_info->nr - 1) / m_numberCaves;

        process_game();
      } else {
        error = -1;   // too high level nr
      }

      if (error == 0) 
      {
        cave_t  *cave = &m_cave; 

        /* convert now */
        itemSteelwall = item_from_name("STEELWALL");

        if(m_numberCaves > 1) {
          level_file_info->packed = TRUE;
        } else {
          level_file_info->packed = FALSE;
        }

        /* check intermission */
        if (cave->intermission) {
          int mx;
          cave->intermissionChecked = TRUE;
          for (y=0; y<m_cavy; y++) {
            if (y < BDCFF_INTERMISSION_HEIGHT) {
              mx = BDCFF_INTERMISSION_WIDTH;
            } else {
              mx = 0;
            }
            for (x=mx; x<m_cavx; x++) {
              if (cave->map_matrix[m_lev][y][x] != itemSteelwall) {
                cave->intermissionChecked = FALSE;
                y = m_cavy+1;
                break;
              }      
            }
          }
        }

        if (cave->intermissionChecked) {
          maxx    = BDCFF_INTERMISSION_WIDTH;
          maxy    = BDCFF_INTERMISSION_HEIGHT;
        } else {
          maxx    = BDCFF_CAVE_WIDTH;
          maxy    = BDCFF_CAVE_HEIGHT;
        }
        level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
        level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy), MAX_LEV_FIELDY);

        for (y=0; y<maxy; y++) {
          for (x=0; x<maxx; x++) {
            level->field[x][y] = cave->map_matrix[m_lev][y][x];   // 1:1 
          }
        }
        /* last line (y=maxy) */
        /*-- not needed in BDCFF
        y = maxy;
        for (x=0; x<maxx; x++) {
          level->field[x][y] = itemSteelwall;   // 1:1
        }
        --*/

        /*-----------------------------------*/
        level->gems_needed              = cave->diamonds_required[m_lev];
        level->score[SC_EMERALD]        = cave->diamond_value;
        level->score[SC_TIME_BONUS]     = cave->bonus_diamond_value;
        level->grow_into_diggable       = TRUE;

        t = cave->cave_time[m_lev];
        t = t + options.bd_caveTimeOffset + t * options.bd_caveTimeGrade/100.0;
        level->time                     = t;

        t = cave->amoeba_time;
        t = t + options.bd_amoebaTimeOffset + t * options.bd_amoebaTimeGrade/100.0;
        level->amoeba_speed             = t;

        t = cave->magic_wall_time;
        t = t + options.bd_magicwallTimeOffset + t * options.bd_magicwallTimeGrade/100.0;
        level->time_magic_wall          = t;
    
        /*-----------------------------------*/
        strcpy(level->name, cave->cave_name);

        if (level->name[0] == 0) {
          strcpy(level->name, level_file_info->basename);
        }
      } 
    }
    return (error == 0);
}
