// ============================================================================
// filesGAMEBD.c
// ----------------------------------------------------------------------------
// (c) 2006-2023 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// loader for GAMEBD game files - can use BDCFF config to convert
// which can lead to infinite possibilities
// ----------------------------------------------------------------------------
// credits to Guido Mersmann who create "Boulderdaesh" 
// ============================================================================

#include <ctype.h>
#include <sys/stat.h>
#include <math.h>

#include "libgame/libgame.h"

#include "files.h"
#include "filesBD.h"
#include "filesBDCFF.h"
#include "filesGAMEBD.h"
#include "init.h"
#include "tools.h"
#include "tape.h"
#include "libgame/rndapi.h"		/*#HAG#VC#*/
#include "libgame/util.h"

// global variables
static char    filename[MAX_FILENAME_LEN] = "";
static int     error        = 0;
static int     warning      = 0;


#define ERR_FILE_NOTFOUND             1
#define ERR_FILE_SEEK                 2
#define ERR_FILE_READ                 3
#define ERR_NO_CAVES                  4
#define ERR_NO_ANCHOR                 5
#define ERR_CAVEHEADER_LEN            6
#define ERR_CAVE_LEN                  7


// Types
typedef unsigned char   byte_t;


// Game types
typedef struct gamebd_s { 
  byte_t    gamebd_magic[6+2];
  byte_t    numberOfLevels[4];
  byte_t    caveCurrent[4];
} gamebd_t;

// -> Anchor:   // 
typedef struct anchorHeader_s {
    byte_t      caveFirstLevelAddr[4];
    byte_t      caveNullLevelAddr[4];
    byte_t      caveLastLevelAddr[4];
    byte_t      editorLen[4];             // 0100 dez
    byte_t      gameEditorText[6*16+4];   
    byte_t      gameFileName[132];
} anchorHeader_t;

typedef struct gameListHeader_s {                   // 
    byte_t      caveNextLevelAddr[4];   // 
    byte_t      caveLastLevelAddr[4];   // 
    byte_t      unknownBytes01[2];      // 00 00                
    byte_t      caveNameAddr[4];        // 
} gameListHeader_t;

typedef struct caveHeader_s {
    byte_t      caveName[32+2];         //  "<name>"  0000
    byte_t      bdCave[6+2];            //  "bdCave"  0000
    byte_t      bdVersion[2];           //  0004
    byte_t      sizex[2];               //  
    byte_t      sizey[2];               //  

    byte_t      headsize[2];            //  0128   // always
    byte_t      levelLength[2];         //  + 2    
    byte_t      intermission[2];        //  + 4    // 0008 normal -> 000C intermission  (bits: 1000 1100) 002A 002E (1010 1110)
    byte_t      magicWallTime[2];       //  + 6    
    byte_t      levelTime[2];           //  + 8    
    byte_t      watergrowingtime[2];    //  +10    
    byte_t      watermaximum[1];        //  +12    
    byte_t      watergrowingrate[1];    //  +13    
    byte_t      diamondsNeeded[2];      //  +14    

    byte_t      unknown01[2];           //  +16    
    byte_t      diamondPoints[2];       //  +18    
    byte_t      unknown02[2];           //  +20    
    byte_t      diamondExtra[2];        //  +22    
    byte_t      unknown03[2];           //  +24    
    byte_t      diamondPoints99[2];     //  +26    
    byte_t      unknown04[2];           //  +28    
    byte_t      forBonusLife[2];        //  +30    

    byte_t      unknown05[2];           //  +32    
    byte_t      timeBonus[2];           //  +34    
    byte_t      gameSpeed[2];           //  +36    
    byte_t      gravity[2];             //  +38    // 0002=left 0004=right 0000=down 0001=up 
    byte_t      waterfloodrate[1];      //  +40    
    byte_t      slimerate[1];           //  +41    
    byte_t      unknown06[2];           //  +42    
    byte_t      unknown07[2];           //  +44    
    byte_t      magicwatertime[2];      //  +46    

    byte_t      earthQuake[2];          //  +48    // 0000=Off 0001=Shuffle 0002=Once 0003=Cycle 
    byte_t      activeTime[2];          //  +50    
    byte_t      earthDelay[2];          //  +52    
    byte_t      earthLength[2];         //  +54    
    byte_t      forceRate[1];           //  +56    
    byte_t      unknown08[1];           //  +57    
    byte_t      unknown09[2];           //  0000
    byte_t      unknown10[2];           //  0000
    byte_t      unknown11[2];           //  0000

    byte_t      unknown12[10];          //  0000 0000 0000 0000 0000

    byte_t      gfxName[32+2];
    byte_t      fntName[32+2];
    byte_t      typeName[32+2];
    byte_t      musName[32+2];
    byte_t      colorSetName[32+2];
    
    byte_t      levelScoreStartAddr[4];
} caveHeader_t;

typedef struct cave_s {
    byte_t      cave[100*100+16+1];
} cave_t;


//-- Game properties -------------------------------------
static long                 m_numberOfLevels = 0;
static int                  m_level          = 0;

static gamebd_t             m_gamebd; 
static anchorHeader_t       m_anchorHeader; 
static gameListHeader_t     m_gameListHeader; 
static caveHeader_t         m_caveHeader; 
static cave_t               m_cave; 

static int                  m_gamebd_init    = 0;

//-- Game elements ---------------------------------------

static int  gamebd_empty               = -1;
static int  gamebd_fakeempty           = -1;
static int  gamebd_diamond             = -1;
static int  gamebd_rock                = -1;
static int  gamebd_fakerock            = -1;

static int  gamebd_egg                 = -1;
static int  gamebd_dirt                = -1;
static int  gamebd_fakedirt            = -1;
static int  gamebd_steelwall           = -1;
static int  gamebd_fakesteel           = -1;

static int  gamebd_wall                = -1;
static int  gamebd_magicwall           = -1;
static int  gamebd_growh               = -1;
static int  gamebd_growv               = -1;
static int  gamebd_water               = -1;

static int  gamebd_firefly_down        = -1;
static int  gamebd_butterfly_up        = -1;
static int  gamebd_playerin            = -1;
static int  gamebd_outvisible          = -1;
static int  gamebd_outhidden           = -1;

static int  gamebd_outfake             = -1;
static int  gamebd_twin                = -1;
static int  gamebd_slime               = -1;
static int  gamebd_gravup              = -1;
static int  gamebd_gravdown            = -1;

static int  gamebd_gravleft            = -1;
static int  gamebd_gravright           = -1;
static int  gamebd_earthquake          = -1;
static int  gamebd_emptyblocked        = -1;

// special
static int  gamebd_diamond_creating    = -1;
static int  gamebd_twin_fake           = -1;
// firefly
static int  gamebd_firefly_right       = -1;
static int  gamebd_firefly_left        = -1;
static int  gamebd_firefly_up          = -1;
// butterfly
static int  gamebd_butterfly_left      = -1;
static int  gamebd_butterfly_down      = -1;
static int  gamebd_butterfly_right     = -1;


/*----------------------------------------------------------------------*/
static void gamebd_init(void)
{
    gamebd_empty               = BDCFF_itemFromName("GAMEBD_EMPTY",             EL_EMPTY                        );
    gamebd_fakeempty           = BDCFF_itemFromName("GAMEBD_FAKEEMPTY",         EL_INVISIBLE_WALL               );
    gamebd_diamond             = BDCFF_itemFromName("GAMEBD_DIAMOND",           EL_BD_DIAMOND                   );
    gamebd_rock                = BDCFF_itemFromName("GAMEBD_ROCK",              EL_BD_ROCK                      );
    gamebd_fakerock            = BDCFF_itemFromName("GAMEBD_FAKEROCK",          EL_ROCK                         );

    gamebd_egg                 = BDCFF_itemFromName("GAMEBD_EGG",               EL_NUT                          );
    gamebd_dirt                = BDCFF_itemFromName("GAMEBD_DIRT",              EL_SAND                         );
    gamebd_fakedirt            = BDCFF_itemFromName("GAMEBD_FAKEDIRT",          EL_INVISIBLE_SAND               );
    gamebd_steelwall           = BDCFF_itemFromName("GAMEBD_STEELWALL",         EL_STEELWALL                    );
    gamebd_fakesteel           = BDCFF_itemFromName("GAMEBD_FAKESTEEL",         EL_INVISIBLE_STEELWALL          );

    gamebd_wall                = BDCFF_itemFromName("GAMEBD_WALL",              EL_BD_WALL                      );
    gamebd_magicwall           = BDCFF_itemFromName("GAMEBD_MAGICWALL",         EL_BD_MAGIC_WALL                );
    gamebd_growh               = BDCFF_itemFromName("GAMEBD_GROWH",             EL_EXPANDABLE_WALL_HORIZONTAL   );
    gamebd_growv               = BDCFF_itemFromName("GAMEBD_GROWV",             EL_EXPANDABLE_WALL_VERTICAL     );
    gamebd_water               = BDCFF_itemFromName("GAMEBD_WATER",             EL_AMOEBA_WET                   );

    gamebd_firefly_down        = BDCFF_itemFromName("GAMEBD_FIREFLY_DOWN",      EL_BD_FIREFLY_DOWN              );
    gamebd_butterfly_up        = BDCFF_itemFromName("GAMEBD_BUTTERFLY_UP",      EL_BD_BUTTERFLY_UP              );
    gamebd_playerin            = BDCFF_itemFromName("GAMEBD_PLAYERIN",          EL_PLAYER_1                     );
    gamebd_outvisible          = BDCFF_itemFromName("GAMEBD_OUTVISIBLE",        EL_EM_EXIT_CLOSED               );
    gamebd_outhidden           = BDCFF_itemFromName("GAMEBD_OUTHIDDEN",         EL_EM_STEEL_EXIT_CLOSED         );

    gamebd_outfake             = BDCFF_itemFromName("GAMEBD_OUTFAKE",           EL_SIGN_EMERGENCY_EXIT          );
    gamebd_twin                = BDCFF_itemFromName("GAMEBD_TWIN",              EL_PLAYER_1                     );
    gamebd_slime               = BDCFF_itemFromName("GAMEBD_SLIME",             EL_BD_AMOEBA                    );
    gamebd_gravup              = BDCFF_itemFromName("GAMEBD_GRAVUP",            EL_SP_GRAVITY_PORT_UP           );
    gamebd_gravdown            = BDCFF_itemFromName("GAMEBD_GRAVDOWN",          EL_SP_GRAVITY_PORT_DOWN         );

    gamebd_gravleft            = BDCFF_itemFromName("GAMEBD_GRAVLEFT",          EL_SP_GRAVITY_PORT_LEFT         );
    gamebd_gravright           = BDCFF_itemFromName("GAMEBD_GRAVRIGHT",         EL_SP_GRAVITY_PORT_RIGHT        );
    gamebd_earthquake          = BDCFF_itemFromName("GAMEBD_EARTHQUAKE",        EL_CUSTOM_1                     );
    gamebd_emptyblocked        = BDCFF_itemFromName("GAMEBD_EMPTYBLOCKED",      EL_EMPTY_SPACE_1                );

    // special
    gamebd_diamond_creating    = BDCFF_itemFromName("GAMEBD_DIAMCREATE",        EL_BD_DIAMOND                   );
    gamebd_twin_fake           = BDCFF_itemFromName("GAMEBD_TWINFAKE",          EL_PLAYER_2                     );
    // firefly 
    gamebd_firefly_right       = BDCFF_itemFromName("GAMEBD_FIREFLY_RIGHT",     EL_BD_FIREFLY_RIGHT             );
    gamebd_firefly_left        = BDCFF_itemFromName("GAMEBD_FIREFLY_LEFT",      EL_BD_FIREFLY_LEFT              );
    gamebd_firefly_up          = BDCFF_itemFromName("GAMEBD_FIREFLY_UP",        EL_BD_FIREFLY_UP                );
    // butterfly 
    gamebd_butterfly_left      = BDCFF_itemFromName("GAMEBD_BUTTERFLY_LEFT",    EL_BD_BUTTERFLY_LEFT            );
    gamebd_butterfly_down      = BDCFF_itemFromName("GAMEBD_BUTTERFLY_DOWN",    EL_BD_BUTTERFLY_DOWN            );
    gamebd_butterfly_right     = BDCFF_itemFromName("GAMEBD_BUTTERFLY_RIGHT",   EL_BD_BUTTERFLY_RIGHT           );
}


/*----------------------------------------------------------------------*/
static int gamebd_map(int x, int y, byte_t ch) 
{
  int  el;
  char elHex[40] = {0};
  sprintf(elHex, "x: %d y: %d ch: %2.2x", x, y, ch);           // x, y

  switch (ch) {
    /*--*/
    case 0x00 : el = gamebd_empty;                  // EL_EMPTY;                       // empty
                break;
    case 0x8B : el = gamebd_fakeempty;              // EL_INVISIBLE_WALL;              // fake empty
                break;
    case 0x80 : el = gamebd_diamond;                // EL_BD_DIAMOND;                  // diamond
                break; 
    case 0x02 : el = gamebd_rock;                   // EL_BD_ROCK;                     // rock
                break;
    case 0x0B : el = gamebd_fakerock;               // EL_ROCK;                        // fake rock
                break;

    case 0x40 : el = gamebd_egg;                    // EL_NUT;                         // egg
                break;
    case 0x01 : el = gamebd_dirt;                   // EL_SAND;                        // dirt
                break;
    case 0x11 : el = gamebd_fakedirt;               // EL_INVISIBLE_SAND;              // fake dirt
                break;
    case 0x04 : el = gamebd_steelwall;              // EL_STEELWALL;                   // steel wall
                break;
    case 0x0A : el = gamebd_fakesteel;              // EL_INVISIBLE_STEELWALL;         // fake metal
                break;

    case 0x05 : el = gamebd_wall;                   // EL_BD_WALL;                     // wall
                break;
    case 0x85 : el = gamebd_magicwall;              // EL_BD_MAGIC_WALL;               // magic wall
                break;
    case 0x06 : el = gamebd_growh;                  // EL_EXPANDABLE_WALL_HORIZONTAL;  // growing wall horizontal
                break;
    case 0x07 : el = gamebd_growv;                  // EL_EXPANDABLE_WALL_VERTICAL;    // growing wall vertical
                break;
    case 0x82 : el = gamebd_water;                  // EL_AMOEBA_WET;                  // water
                break;

    case 0xE2 : el = gamebd_firefly_down;           // EL_BD_FIREFLY_DOWN;             // firefly (down)
                break;
    case 0xE5 : el = gamebd_butterfly_up;           // EL_BD_BUTTERFLY_UP;             // butterfly (up)
                break;
    case 0x86 : el = gamebd_playerin;               // EL_PLAYER_1;                    // player in
                break;
    case 0x87 : el = gamebd_outvisible;             // EL_EM_EXIT_CLOSED;              // visible out
                break;
    case 0x08 : el = gamebd_outhidden;              // EL_EM_STEEL_EXIT_CLOSED;        // hidden out
                break;

    case 0x88 : el = gamebd_outfake;                // EL_SIGN_EMERGENCY_EXIT;         // visible out fake exit
                break;
    case 0x09 : el = gamebd_twin;                   // EL_PLAYER_1;                    // twin
                break;
    case 0x83 : el = gamebd_slime;                  // EL_BD_AMOEBA;                   // slime
                break;
    case 0x0C : el = gamebd_gravup;                 // EL_SP_GRAVITY_PORT_UP;          // gravity up
                break;
    case 0x0D : el = gamebd_gravdown;               // EL_SP_GRAVITY_PORT_DOWN;        // gravity down
                break;

    case 0x0E : el = gamebd_gravleft;               // EL_SP_GRAVITY_PORT_LEFT;        // gravity left
                break;
    case 0x0F : el = gamebd_gravright;              // EL_SP_GRAVITY_PORT_RIGHT;       // gravity right
                break;
    case 0x8A : el = gamebd_earthquake;             // earthquake
                break;
    case 0x10 : el = gamebd_emptyblocked;           // block empty
                break;

    // special
    case 0x38 : el = gamebd_diamond_creating;       // EL_BD_DIAMOND                   // diamond creating
                break;

    case 0x89 : el = gamebd_twin_fake;              // EL_PLAYER_1                     // twin fake
                break;
     
    // firefly
    case 0xE0 : el = gamebd_firefly_right;          // EL_BD_FIREFLY_RIGHT             // firefly (right)
                break;
    case 0xE1 : el = gamebd_firefly_left;           // EL_BD_FIREFLY_LEFT              // firefly (left)
                break;
    case 0xB2 : el = gamebd_firefly_up;             // EL_BD_FIREFLY_UP                // firefly (up)
                break;

    // butterfly
    case 0xE3 : el = gamebd_butterfly_left;         // EL_BD_BUTTERFLY_LEFT;           // butterfly (left)
                break;
    case 0xE4 : el = gamebd_butterfly_down;         // EL_BD_BUTTERFLY_DOWN;           // butterfly (down)
                break;
    case 0xB3 : el = gamebd_butterfly_right;        // EL_BD_BUTTERFLY_RIGHT           // butterfly (right)
                break;

    default   : sprintf(elHex, "x: %d y: %d ch: %2.2x", x, y, ch);           // x, y
                el = EL_STEEL_CHAR_QUESTION;
                break;
  }
  return el;
}


/*----------------------------------------------------------------------*/
static long convBytes2offset (byte_t *what, int len) {
    long result = 0;
    long multi  = 1;
    for (int i=len-1; i>=0; i--) {
        result += what[i] * multi;
        multi *= 256;
    }
    return result;
}

/*----------------------------------------------------------------------*/
static long bytes2offset (byte_t *what) {
    return convBytes2offset(what, 4);
}

/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
boolean Load_GAMEBDlevel(struct LevelInfo *level,
                         struct LevelFileInfo *level_file_info)
{
    File        *file;
    int         fok = 0;
    size_t      rd  = 0;

    char        textGameEditor[] = "Game Editor";

    long        fileOffset   = 0;
    long        anchorOffset = 0x10;
    long        anchorNull   = anchorOffset + 4;

    int         actlev = 0;
    byte_t      headerOffsetBytes[4] = {0};

    // working
    error        = 0;
    warning      = 0;

    // init
    if (m_gamebd_init == FALSE) {
        BDCFF_init();
        gamebd_init();
        m_gamebd_init = TRUE;
    }

    // init
    m_numberOfLevels = 0;
    m_level          = 0;
    memset(&m_gamebd,         0, sizeof(m_gamebd));
    memset(&m_anchorHeader,   0, sizeof(m_anchorHeader));
    memset(&m_gameListHeader, 0, sizeof(m_gameListHeader));
    memset(&m_caveHeader,     0, sizeof(m_caveHeader));
    memset(&m_cave,           0, sizeof(m_cave));

    strcpy (filename, level_file_info->filename);
    file = openFile(filename, "rb");
    if (file != NULL) {
        // read gamebd
        if(error == 0) {
            fok = seekFile(file, fileOffset, SEEK_SET);
            if (fok == 0) {
                rd = readFile(file, &m_gamebd, sizeof(gamebd_t), 1);
                if (!(rd > 0)) {
                    error = ERR_FILE_READ;
                }
            } else {
                error = ERR_FILE_SEEK;
            }
        }
        // check
        if(error == 0) {
            m_numberOfLevels = bytes2offset(m_gamebd.numberOfLevels);
            if (m_numberOfLevels <= 0) {
                error = ERR_NO_CAVES;
            }
        }
        if(error == 0) {
            if(m_numberOfLevels > 1) {
              level_file_info->packed = TRUE;
            } else {
              level_file_info->packed = FALSE;
            }

            if (level_file_info->nr <= m_numberOfLevels) {
                m_level = (level_file_info->nr - 1);
            } else {
                error = -1;   // too high level nr
            }
        }
        // read anchor
        if(error == 0) {
            // -- anchor header
            fok = seekFile(file, anchorOffset, SEEK_SET);
            if (fok == 0) {
                rd = readFile(file, &m_anchorHeader, sizeof(anchorHeader_t), 1);
                if (rd > 0) {
                    char *cmpPtr = strstr((const char*) &m_anchorHeader.gameEditorText[0], textGameEditor);
                    if (cmpPtr != (const char*) &m_anchorHeader.gameEditorText[0]) {
                        error = ERR_NO_ANCHOR;
                    } else {
                        // ok
                    }
                }
            }
        }
        if(error == 0) {
            anchorNull = anchorOffset + 4;
        }

        // go to the level
        if(error == 0) {
            memcpy(&headerOffsetBytes[0], &m_anchorHeader.caveFirstLevelAddr[0], 4);

            while(1) {
                // -- header (from list)
                long headeroffset = bytes2offset(headerOffsetBytes);

                if (headeroffset == anchorNull) {
                    if (actlev < m_level) {
                        error = -1;  // too high level nr
                    }
                    break;
                }

                fok = seekFile(file, headeroffset, SEEK_SET);
                if (fok == 0) {
                    rd = readFile(file, &m_gameListHeader, sizeof(gameListHeader_t), 1);
                    if (!(rd > 0)) {
                        error = ERR_FILE_READ;
                        break;
                    }
                }

                if (actlev == m_level) {
                    break;  // found
                }
                // copy caveNextLevelAddr to headerOffsetBytes
                memcpy(&headerOffsetBytes[0], &m_gameListHeader.caveNextLevelAddr[0], 4);
                // next level
                actlev++;
            }
        }

        if(error == 0) {
            // -- cave header (from name)
            long caveoffset = bytes2offset(&m_gameListHeader.caveNameAddr[0]);

            fok = seekFile(file, caveoffset, SEEK_SET);
            if (fok == 0) {
                rd = readFile(file, &m_caveHeader, sizeof(caveHeader_t), 1);
                if (!(rd > 0)) {
                    error = ERR_FILE_READ;
                }
            }
        }

        if(error == 0) {
            // -- cave
            long addrfrom      = bytes2offset(&m_gameListHeader.caveNameAddr[0]);
            long caveheaderlen = convBytes2offset(&m_caveHeader.headsize[0], sizeof(m_caveHeader.headsize));
            if (caveheaderlen != 0x0128) {
                error = ERR_CAVEHEADER_LEN;
            }
            long addrlev  = addrfrom + 0x128;
            long cavelen  = convBytes2offset(&m_caveHeader.levelLength[0], 2);

            fok = seekFile(file, addrlev, SEEK_SET);
            if (fok == 0) {
                long   index   = 0;
                while(1) {
                    byte_t levb = 0;
                    rd = readFile(file, &levb, 1, 1);
                    if (rd > 0) {
                        m_cave.cave[index] = levb;
                        index++;
                        if (levb == 0xFF) {
                            // read element
                            rd = readFile(file, &levb, 1, 1);
                            if (rd > 0) {
                                m_cave.cave[index] = levb;
                                index++;
                            }
                            if (levb == 0xFF) {
                                break;
                            }
                            // read numberOf
                            rd = readFile(file, &levb, 1, 1);
                            if (rd > 0) {
                                m_cave.cave[index] = levb;
                                index++;
                            }
                        }
                    }
                }
                // check length
                if (index != cavelen) {
                    error = ERR_CAVE_LEN;
                }
            }
        }

        // all data collected
        closeFile(file);

        if(error == 0) {
            int maxx = convBytes2offset(m_caveHeader.sizex, sizeof(m_caveHeader.sizex));
            int maxy = convBytes2offset(m_caveHeader.sizey, sizeof(m_caveHeader.sizey));
            int cavelen = convBytes2offset(m_caveHeader.levelLength, sizeof(m_caveHeader.levelLength));
            int x    = 0;
            int y    = 0;
            //
            level->fieldx = MIN(MAX(MIN_LEV_FIELDX, maxx), MAX_LEV_FIELDX);
            level->fieldy = MIN(MAX(MIN_LEV_FIELDY, maxy), MAX_LEV_FIELDY);
    
            for(int i=0; i < cavelen; i++) {
                byte_t levb = m_cave.cave[i];
                byte_t elem = levb;
                byte_t cnte = 1;
                int    sele = 0;
                if (levb == 0xFF) {
                    i++;
                    elem = m_cave.cave[i];
                    if (elem == 0xFF) {
                        if (i+1 != cavelen) {
                            error = ERR_CAVE_LEN;
                        }
                        break;
                    }
                    i++;
                    cnte = m_cave.cave[i];
                }
                sele = gamebd_map(x, y, elem);  // convert
                if (sele == EL_STEEL_CHAR_QUESTION) {
                    sele = gamebd_map(x, y, elem);  // convert
                }
                for(int e=0; e < cnte; e++) {
                    level->field[x][y] = sele;
                    x++;
                    if (x >= maxx) {
                        x = 0;
                        y++;
                    }
                }
            }
        }

        if(error == 0) {
            int t;
            /*-----------------------------------*/
            level->gems_needed              = convBytes2offset(m_caveHeader.diamondsNeeded, 2);
            level->score[SC_EMERALD]        = convBytes2offset(m_caveHeader.diamondPoints, 2);
            level->score[SC_TIME_BONUS]     = convBytes2offset(m_caveHeader.diamondExtra,2);
            level->grow_into_diggable       = TRUE;

            t = convBytes2offset(m_caveHeader.levelTime, 2);
            t = t + options.bd_caveTimeOffset + t * options.bd_caveTimeGrade/100.0;
            level->time                     = t;

            t = convBytes2offset(m_caveHeader.watergrowingtime, 2);
            t = t + options.bd_amoebaTimeOffset + t * options.bd_amoebaTimeGrade/100.0;
            level->amoeba_speed             = t;

            t = convBytes2offset(m_caveHeader.magicWallTime, 2);
            t = t + options.bd_magicwallTimeOffset + t * options.bd_magicwallTimeGrade/100.0;
            level->time_magic_wall          = t;
    
            /*-----------------------------------*/
            strcpy(level->name, (const char*)&m_caveHeader.caveName[0]);

            if (level->name[0] == 0) {
              strcpy(level->name, level_file_info->basename);
            }

        }
    } else {
        error = ERR_FILE_NOTFOUND;
    }
    return (error == 0);;
}

