// ============================================================================
// mouse.c
// ----------------------------------------------------------------------------
// (c) 2006-2023 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// handle mouse option and mouse.conf
// ============================================================================

#include <time.h>
#include <sys/types.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>

#include "platform.h"

#if !defined(PLATFORM_WINDOWS)
#include <pwd.h>
#include <sys/param.h>
#include <sys/time.h>		/*#HAG#VC#*/
#include <unistd.h>		/*#HAG#VC#*/
#else
#include "rndapi.h"		/*#HAG#VC#*/
#endif

#include "mouse.h"
#include "setup.h"
#include "util.h"


/*--#HAG#mouse#--*/
/*-----------------------------------------------------------------------
 calcMouseOption - calculate internal mouse options 
-----------------------------------------------------------------------*/
void calcMouseOption (long i_mode) 
{
  /* init other   */
  options.mouseOption.learn             = FALSE;
  options.mouseOption.learnPause        = FALSE;      
  options.mouseOption.directMove        = FALSE;  
  options.mouseOption.sokoMode          = 0;      
  options.mouseOption.snokoMode         = 0;      
  /* init control */
  options.mouseOption.frameDelay        = 20;

  switch (i_mode) {
    case MOUSEMODE_SOKO_SNOKO1 :   /* 1 */ /* LEFT = SOKO, RIGHT = SNOKO */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = FALSE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_SOKO_SNOKO2 :   /* 2 */ /* LEFT = SOKO, RIGHT = SNOKO */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_SMALL;      
      break;
    case MOUSEMODE_SOKO_SNOKO3 :   /* 3 */ /* LEFT = SOKO, RIGHT = SNOKO */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;

    case MOUSEMODE_FOLLOW1     :   /* 11 *//* RIGHT = with follow mouse */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = TRUE;   /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = FALSE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_FOLLOW2     :   /* 12 *//* RIGHT = with follow mouse */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = TRUE;   /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_SMALL;      
      break;
    case MOUSEMODE_FOLLOW3     :   /* 13 *//* RIGHT = with follow mouse */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = TRUE;   /* can toggle! */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SOKO_SNOKO;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;

    case MOUSEMODE_FOLLOWSNAP1 :   /* 21 *//* with SNAP */ 
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = MHM_SNOKO;
      options.mouseOption.mouseHandleButton = MHB_DIRSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = FALSE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_FOLLOWSNAP2 :  /* 22 *//* with SNAP */ 
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = MHM_SNOKO;
      options.mouseOption.mouseHandleButton = MHB_DIRSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_SMALL;      
      break;
    case MOUSEMODE_FOLLOWSNAP3 :  /* 23 *//* with SNAP */ 
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = MHM_SNOKO;
      options.mouseOption.mouseHandleButton = MHB_DIRSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;

    case MOUSEMODE_DIRDIFF     :  /* 31 *//* direct follow mouse */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = MHM_DIRDIFF;
      options.mouseOption.mouseHandleButton = MHB_DIRSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;  /* not used here! */    
      options.mouseOption.snokoMode         = 0;  /* not used here! */   
      break;

    case MOUSEMODE_SNOKOSNAP1 :   /* 41 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = FALSE;   /* snap can force this! */
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_SNOKOSNAP2 :   /* 42 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = TRUE;      
      options.mouseOption.directMove        = FALSE;   /* snap can force this! */  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_SNOKOSNAP3 :   /* 43 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_SMALL;      
      break;
    case MOUSEMODE_SNOKOSNAP4 :   /* 44 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user*/
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = TRUE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_SMALL;      
      break;
    case MOUSEMODE_SNOKOSNAP5 :   /* 45 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;
    case MOUSEMODE_SNOKOSNAP6 :   /* 46 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = TRUE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;
    case MOUSEMODE_SNOKOSNAP7 :   /* 47 = 41 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = FALSE;      
      options.mouseOption.directMove        = TRUE;   /* snap can force this! */
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_SNOKOSNAP8 :   /* 48 = 42 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = TRUE;      
      options.mouseOption.directMove        = TRUE;   /* snap can force this! */  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = 0;      
      break;
    case MOUSEMODE_SNOKOSNAP9 :   /* 49 = 46 *//* snoko snap */
      options.mouseOption.mode              = i_mode;
      options.mouseOption.followMouse       = FALSE;  /* toggle by user */
      options.mouseOption.mouseHandleMove   = 0;
      options.mouseOption.mouseHandleButton = MHB_SNOKOSNAP;
      options.mouseOption.learn             = TRUE;      
      options.mouseOption.learnPause        = TRUE;      
      options.mouseOption.directMove        = TRUE;  
      options.mouseOption.sokoMode          = 0;      
      options.mouseOption.snokoMode         = SNOKOMODE_DIRECT;      
      break;

    case MOUSEMODE_OFF :
    default            :
      options.mouseOption.mode              = 0;      
      options.mouseOption.followMouse       = FALSE;  
      /* MouseHandleMove   */
      options.mouseOption.mouseHandleMove   = 0;      /* MHM_ */
      /* MouseHandleButton */
      options.mouseOption.mouseHandleButton = 0;      /* MHB_ */
      /* other */
      options.mouseOption.directMove        = FALSE;  /* move is not calculated      */
      options.mouseOption.sokoMode          = 0;      /* calculation mode SOKOMODE_  */
      options.mouseOption.snokoMode         = 0;      /* calculation mode SNOKOMODE_ */
      break;
  }
}


/* myGetPath2 - build path from two  
*/  
static char *myGetPath2(char *path1, char *path2)
{
  char *complete_path = objectNew(strlen(path1) + 1 +
				  strlen(path2) + 1);
  sprintf(complete_path, "%s/%s", path1, path2);
  return complete_path;
}

/* myGetPath2ext - build path from two with ext 
*/  
static char *myGetPath2ext(char *path1, char *path2, char *ext)
{
  char *complete_path = objectNew(strlen(path1) + 1 +
				  strlen(path2) + 1 +
                                  strlen(ext)   + 1);
  sprintf(complete_path, "%s/%s%s", path1, path2, ext);
  return complete_path;
}

/*-----------------------------------------------------------------------
 SetupMouseOptionFromConf 
-----------------------------------------------------------------------*/
void SetupMouseOptionFromConf (char *basedir, char* filename)
{
  char *confFilename = NULL;
  SetupFileHash *file_setup_hash;

  // get config 
  if (strchr(filename, '.')) {
    confFilename = myGetPath2(basedir, filename);
  } else {
    confFilename = myGetPath2ext(basedir, filename, ".conf");
  }
  if ((file_setup_hash = loadSetupFileHash(confFilename)))
  {
    char *token_value;

    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_MODE);
    if (token_value) options.mouseOption.mode               = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_FOLLOWMOUSE);
    if (token_value) options.mouseOption.followMouse        = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_MOUSEHANDLEMOVE);
    if (token_value) options.mouseOption.mouseHandleMove    = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_MOUSEHANDLEBUTTON);
    if (token_value) options.mouseOption.mouseHandleButton  = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_LEARN);
    if (token_value) options.mouseOption.learn              = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_LEARNPAUSE);
    if (token_value) options.mouseOption.learnPause         = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_DIRECTMOVE);
    if (token_value) options.mouseOption.directMove         = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_SOKOMODE);
    if (token_value) options.mouseOption.sokoMode           = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_SNOKOMODE);
    if (token_value) options.mouseOption.snokoMode          = atoi(token_value);
    token_value = getHashEntry(file_setup_hash, MOUSE_TOKEN_FRAMEDELAY);
    if (token_value) options.mouseOption.frameDelay         = atoi(token_value);

    /*-- not needed -->
    checkSetupFileHashIdentifier(file_setup_hash, confFilename, getCookie("MOUSESETUP"));
    <--*/
    /*-->
    printf("options.mouseOption.mode: %d\n", options.mouseOption.mode);
    printf("options.mouseOption.followMouse: %d\n", options.mouseOption.followMouse);
    printf("options.mouseOption.mouseHandleMove: %d\n", options.mouseOption.mouseHandleMove);
    printf("options.mouseOption.mouseHandleButton: %d\n", options.mouseOption.mouseHandleButton);
    printf("options.mouseOption.learn: %d\n", options.mouseOption.learn);
    printf("options.mouseOption.learnPause: %d\n", options.mouseOption.learnPause);
    printf("options.mouseOption.directMove: %d\n", options.mouseOption.directMove);
    printf("options.mouseOption.sokoMode: %d\n", options.mouseOption.sokoMode);
    printf("options.mouseOption.snokoMode: %d\n", options.mouseOption.snokoMode);
    printf("options.mouseOption.frameDelay: %d\n", options.mouseOption.frameDelay);
    <--*/

    freeSetupFileHash(file_setup_hash);
  }
  objectDelete((Pointer_t)&confFilename);
}

